-module(egg_types).

-export([samples/3]).
-export([t_to_gen/1, t_to_pattern_gen/1]).
-export([is_subtype/2]).

-include_lib("referl_qc/include/prop_based_testing.hrl").

-ifdef(otp17).
-include_lib("../include/types_r17.hrl").
-else.
-include_lib("../include/types.hrl").
-endif.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% For Paraphrase

samples(Ts, N, S) ->
    Generators = [t_to_gen(T) || T <- Ts],
    Tab = ets:new(testdata, [public, duplicate_bag]),
    pick_and_store(noshrink(Generators), Tab, N, S),
    Result = [Data || {value, Data} <- ets:lookup(Tab, value)],
    true = ets:delete(Tab),
    Result.

pick_and_store(G, Tab, N, S) ->
    ?QuickCheck:quickcheck(?QuickCheck:numtests(N, ?FORALL(X, resize(S, G), ets:insert(Tab, {value, X})))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



is_subtype(?union(Types), T2) ->
    lists:any(fun(T) -> is_subtype(T, T2) end,
              lists:filter(fun(T) ->  T =/= ?none end, Types));
is_subtype(T1, T2) -> erl_types:t_is_subtype(T1, T2) orelse erl_types:t_is_subtype(T2, T1).



%%=============================================================================
%%
%% EQC generator: type description -> Erlang term
%%
%%=============================================================================

t_to_gen(L) when is_list(L) ->
    lists:map(fun t_to_gen/1, L);
t_to_gen(?any) ->
    42;
t_to_gen(?none) ->
    42;
t_to_gen(?unit) ->
    42;
t_to_gen(?atom(?any)) ->
    [choose($a, $z) | list(choose($a, $z))];
t_to_gen(?atom(Set)) ->
    elements(set_to_list(Set));

%% t_to_gen(?bitstr(8, 0), _RecDict) ->
%%   "binary()";
%% t_to_gen(?bitstr(0, 0), _RecDict) ->
%%   "<<>>";
%% t_to_gen(?bitstr(0, B), _RecDict) ->
%%   io_lib:format("<<_:~w>>", [B]);
%% t_to_gen(?bitstr(U, 0), _RecDict) ->
%%   io_lib:format("<<_:_*~w>>", [U]);
%% t_to_gen(?bitstr(U, B), _RecDict) ->
%%   io_lib:format("<<_:~w,_:_*~w>>", [B, U]);

%% t_to_gen(?function(?any, ?any), _RecDict) ->
%%   "fun()";
%% t_to_gen(?function(?any, Range), RecDict) ->
%%   "fun((...) -> " ++ t_to_gen(Range, RecDict) ++ ")";
%% t_to_gen(?function(?product(ArgList), Range), RecDict) ->
%%   "fun((" ++ comma_sequence(ArgList, RecDict) ++ ") -> "
%%     ++ t_to_gen(Range, RecDict) ++ ")";

t_to_gen(?identifier(Set)) ->
    if Set =:= ?any -> self(); %% :)
       true -> elements(set_to_list(Set))
    end;
t_to_gen(?opaque([#opaque{mod=dict}])) ->
    dict:new();
%    t_to_gen(Struct);
t_to_gen(?opaque(_Set)) ->
    42;
%% t_to_gen(?opaque(Set), _RecDict) ->
%%   sequence([case is_opaque_builtin(Mod, Name) of
%% 	      true  -> io_lib:format("~w()", [Name]);
%% 	      false -> io_lib:format("~w:~w()", [Mod, Name])
%% 	    end
%% 	    || #opaque{mod = Mod, name = Name} <- set_to_list(Set)], [], " | ");
%% t_to_gen(?matchstate(Pres, Slots), RecDict) ->
%%   io_lib:format("ms(~s,~s)", [t_to_gen(Pres, RecDict),
%% 			      t_to_gen(Slots,RecDict)]);

t_to_gen(?nil) ->
    [];
t_to_gen(?nonempty_list(Contents, Termination)) ->
    [list(t_to_gen(Contents))|t_to_gen(Termination)];
t_to_gen(?list(Contents, ?nil, ?unknown_qual)) ->
    list(t_to_gen(Contents));
t_to_gen(?list(Contents, Termination, ?unknown_qual)) ->
    [list(t_to_gen(Contents))|t_to_gen(Termination)];

t_to_gen(?int_set(Set)) ->
    elements(Set);
%% t_to_gen(?byte) -> "byte()";
%% t_to_gen(?char) -> "char()";
%% t_to_gen(?integer_pos) -> "pos_integer()";
%% t_to_gen(?integer_non_neg) -> "non_neg_integer()";
%% t_to_gen(?integer_neg) -> "neg_integer()";
t_to_gen(?int_range(neg_inf, pos_inf)) ->
    int();
t_to_gen(?int_range(From, pos_inf)) ->
    ?SUCHTHAT(I, int(), I >= From);
t_to_gen(?int_range(neg_inf, To)) ->
    ?SUCHTHAT(I, int(), I =< To);
t_to_gen(?int_range(From, To)) ->
    choose(From, To);
t_to_gen(?integer(?any)) ->
    int();
t_to_gen(?float) ->
    real();
t_to_gen(?number(?any, ?unknown_qual)) ->
    int();

t_to_gen(?product(_List)) ->
  throw("product unimplemented");

t_to_gen(?tuple(?any, ?any, ?any)) ->
  {};

t_to_gen(?tuple(Elements, _Arity, ?any)) ->
  ?LET(L, lists:map(fun t_to_gen/1, Elements), list_to_tuple(L));

t_to_gen(?tuple(Elements, Arity, _Tag)) ->
  t_to_gen(?tuple(Elements, Arity, ?any));

t_to_gen(?tuple_set(_) = S) ->
  oneof([t_to_gen(T) || T <- erl_types:t_tuple_subtypes(S), T =/= ?none]);

t_to_gen(?union(Types)) ->
  oneof([t_to_gen(T) || T <- Types, T =/= ?none]);

t_to_gen(T) ->
  io:format("t_to_gen:~p~n", [T]),
  throw(type_not_supported).



%%=============================================================================
%%
%% EQC generator: type description -> Erlang pattern
%%
%%=============================================================================

varname() ->
    ?LET({C, Cs},
         {choose($A, $Z), list(oneof([choose($a, $z), choose($A, $Z)]))},
         [C|Cs]).

val_or_var(Gen) ->
    oneof([Gen, {variable, varname()}]).

t_to_pattern_gen(L) when is_list(L) ->
    lists:map(fun t_to_pattern_gen/1, L);

t_to_pattern_gen(?any) ->
  {variable, varname()};

t_to_pattern_gen(?list(Contents, ?nil, ?unknown_qual)) ->
    list(t_to_pattern_gen(Contents));

t_to_pattern_gen(?tuple(Elements, _Arity, ?any)) ->
    ?LET(L, lists:map(fun t_to_pattern_gen/1, Elements), list_to_tuple(L));

t_to_pattern_gen(?tuple(Elements, Arity, _Tag)) ->
    t_to_pattern_gen(?tuple(Elements, Arity, ?any));

t_to_pattern_gen(?tuple_set(_) = S) ->
    oneof([t_to_pattern_gen(T) || T <- erl_types:t_tuple_subtypes(S), T =/= ?none]);

t_to_pattern_gen(?union(Types)) ->
    oneof([{variable, varname()} | [t_to_pattern_gen(T) || T <- Types, T =/= ?none]]);

t_to_pattern_gen(X) ->
    val_or_var(t_to_gen(X)).

set_to_list(Set) ->
  ordsets:to_list(Set).
