%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% ===========================================================================
%%% Data structures

-record(attrib,    {num_of_workers :: non_neg_integer(),
                    nw_sum         :: non_neg_integer(),
                    tseqcpu        :: float(),
                    tseqgpu        :: float(),
                    tparcpu        :: float(),
                    tpargpu        :: float(),
                    tseqother  = 0 :: float(),
                    list_length = 0:: non_neg_integer(),
                    label          :: nonempty_string(),
                    metric         :: non_neg_integer()}).

-record(comp,      {node_id        :: gnode(),
                    label          :: nonempty_string(),
                    type           :: skel_type() | {none, cand_type()},
                    attribs        :: #attrib{}}).

-record(tr_seq,    {attribs        :: #attrib{},
                   comps           :: [#comp{}],
                   label           :: nonempty_string(),
                   node_id         :: gnode(),
                   % TODO: if the representation is clear,
                   % give a proper type.
                   structure       :: any()}).

-record(farm, {
          node      :: gnode(),
          job       :: {gnode(), [candidate()]},
          type      :: cand_type(),
          attribs   :: #attrib{}
         }).

-record(pool, {
          node      :: gnode(),
          job       :: {gnode(), [candidate()]},
          type      :: cand_type(),
          embeds    :: [candidate()],
	  attribs   :: #attrib{}
         }).


-record(pipe, {
          nodes     :: [gnode()],
          stages    :: [{gnode() | {transformed, gnode()}, [candidate()]}],
          type      :: cand_type(),
                   %% tuple stands for #farm{} and #pipe{}
          attribs   :: #attrib{}
         }).

-record(dnc, {
          node      :: [gnode()],
          type      :: cand_type(),
          embeds    :: [candidate()],
          attribs   :: #attrib{}
         }).

-record(seq, {
          node        :: gnode(),
          parent_type :: cand_type(),
          attribs     :: #attrib{}
         }).

%%% Type definitions

-type skel_type() :: farm | pipe | dnc | pool.

-type cand_type() :: func | listcomp | mapapp.

-type candidate() :: #farm{} | #pipe{} | #pool{} | #dnc{}.

-type gnode() :: {'$gn', atom(), integer()}.

-type elem() :: {type, cand_type()} | {job | stages,
                                      [{orig, gnode()} | {shaped, gnode()}]}.

-type search_options() :: [{modules, [atom() | gnode()]} |
                           {cand_types, [skel_type()]} |
                           {chains, true | false} |
                           {rewrite, true| false} |
			   {num_of_cores, non_neg_integer()} |
			   {default_length_of_stream, non_neg_integer()} |
			   {filter_benchmark_errors, true | false}].


-define(out(S),    ?out(S, [])).
-ifdef(development_mode).
-define(out(S, A), io:format(S ++ "~n", A)).
-define(silent(E), E).
-else.
-define(out(S, A), no_debug_in_release_mode).
-define(silent(E), begin
                       {ok, F} = file:open("/dev/null", [write]),
                       G = group_leader(),
                       group_leader(F, self()),
                       E,
                       group_leader(G, self())
                   end).
-endif.
