%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Melinda Toth <tothmelinda@elte.hu>
%%% @author Istvan Bozo <bozoistvan@elte.hu>
%%% @doc
%%% <b>API module.</b>
%%%
%%% @todo doc!
-module(refpp_api).
-include("referl_paraphrase.hrl").
-export([run/0, run/1]).

%% @doc The function applies the run/1 function with default
%% values. Every candidate is displayed regardless of its module or
%% type and does not examine embeddings.
-spec run() -> ok.
run() ->
    run([]).

%% @doc The input is a proplist with different options. The function
%% displays candidates from the provided modules with the given types.
%% Options are: {modules, Mods}, {cand_types, CandTypes}, {chains,
%% BoolValue}, {rewrite, BoolValue}. The void options are substituted
%% with default values.  Default values: modules - entire graph,
%% cand_types - every type, chains - false, rewrite - false.

-spec run(PropList :: search_options()) -> ok.
run(PropList) when is_list(PropList)->
    CandList =
        case proplists:get_value(chains, PropList, false) of
            true ->
                refpp_skeleton:get_candidate_chain_combinations(PropList);
             _ ->
                refpp_skeleton:get_candidates(PropList)
        end,
    %SCs = lists:filter(fun(#pool{embeds=[_|_]}) -> false;
    %                      (_) -> true
    %                   end, CandList),
    OutputData = refpp_costs:calc(CandList, PropList),
    spawn(fun()->refpp_output:display(OutputData) end);
run(Arg) ->
    io:format("The argument is not a proplist : ~p~n", [Arg]).

