%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>Provided services</b>
%%% <b>It's important to keep the encoding of this file UTF-8!</b>
%%% @todo doc!

-module(refpp_web_services).

-include("referl_paraphrase.hrl").

% Interface functions to query informations about the registered services
-export([services/0, service/1]).

% Inets httpd callbacks
-export([datasource/3, user_input_handler/3, output/3]).

-define(HTML_TEMPLATE_PATH,
        filename:join([proplists:get_value(server_dir, ?WebSrv:server_info()),
                       "output.html"])).

%%% ============================================================================
%%% Interface

services()->
    case proplists:get_value(url, ?WebSrv:server_info()) of
        undefined -> [];
        Url -> [{output, [{url, string:join([Url, "output"], "/")}]},
                {datasource, [{url, string:join([Url, "datasource"], "/")}]},
                {user_input_handler, [{url, string:join([Url, "user_input_handler"], "/")}]}]
    end.

service(Service)->
    proplists:get_value(Service, services(), []).


output(SessionID, _Env, _Input) ->
    %inform PC Browser that a user has requested to download the page
    refpp_output:user_launched_session(),
    {ok, Content} = file:read_file(?HTML_TEMPLATE_PATH),
    deliver(SessionID, html, [Content]).

datasource(SessionID, _Env, Input) ->
    Parsed = httpd:parse_query(Input),
    {OutputType, Content} = refpp_datasource:handle(Parsed),
    deliver(SessionID, OutputType, Content).

user_input_handler(SessionID, _Env, Input) ->
    try 
        Parsed = httpd:parse_query(Input),
        %?d(Parsed),
        case proplists:get_value("selected_tr_seq", Parsed, not_found) of
            TrSeqStr when is_list(TrSeqStr), length(TrSeqStr)>0 ->
                refpp_output:save_user_req(list_to_integer(TrSeqStr)),
                deliver(SessionID, html, ["<html><body>ok</body></html>"]);
            _ ->
                case proplists:get_value("status", Parsed, not_found) of
                    "unloaded" ->
                        %inform PC Browser that the session is aborted
                        refpp_output:user_aborted_session(),
                        deliver(SessionID, html, ["<html><body>ok</body></html>"]);
                    _ ->
                        % unknown req.
                        refpp_output:user_aborted_session(),
                        deliver(SessionID, html, ["<html><body>ok</body></html>"])
                end
        end
    catch
        Cl:E ->
            % internal error occurred
            refpp_output:user_aborted_session(),
            throw({Cl,E})
    end.

%%% ============================================================================
%%% Internals

deliver(SessionID, ContentType, Content)->
    mod_esi:deliver(
        SessionID,
        headers(ContentType) ++ Content).

headers(xml)->
         ["cache-control: private, max-age=0\n",
         "content-disposition: attachment; filename=result.xml\n",
         "content-type: application/xml; charset=UTF-8\r\n\r\n"];
headers(html)->
         ["cache-control: private, max-age=0\n",
         "content-type: text/html; charset=UTF-8\r\n\r\n"];
headers(csv)->
         ["cache-control: private, max-age=0\n",
          "content-disposition: attachment; filename=result.csv\n",
          "content-type: text/csv; charset=UTF-8\r\n\r\n"];
headers(json)->
        ["cache-control: private, max-age=0\n",
         "content-type: application/json; charset=UTF-8\r\n\r\n"].
