-module(refusr_sq_lib_tests).

-include_lib("eunit/include/eunit.hrl").

-include_lib("referl_gen/include/gen_export.hrl").
-include_lib("referl_core/include/core_global.hrl").
-include_lib("referl_lib/include/lib_export.hrl").


-include_lib("referl_core/include/core_export.hrl").

-export([run_nodes/1]).

-define(Lib, refusr_sq_lib).

setop_testfile1() ->
    {module, ?MISC:canonical_filename("m1.erl"),
    "-module(m1).\n"
    "-export([f/1]).\n"
    "f(Var)-> [Var, m2:h()].\n"}.
    
setop_testfile2() ->
    {module, ?MISC:canonical_filename("m2.erl"),
    "-module(m2).\n"
    "-export([h/0]).\n"
    "h()-> ok.\n"
    "g(stop)-> ok;\n"
    "g(X)-> g(X-1).\n"}.

setop_setup() ->
    [add_file(setop_testfile1()),add_file(setop_testfile2())].
setop_cleanup(Files) ->
    ?debugVal([?FileMan,Files]),[?FileMan:drop_file(File)||File<-Files].

setop_test() ->
    {setup,
        setop_setup(),
        fun setop_cleanup/1,
        [
        ?assertEqual(
            [{group_by,{nopos,"m1:f/1"},eq,name,f},
             {group_by,{nopos,"m2:h/0"},eq,name,h},
             {group_by,{nopos,"m2:g/1"},eq,name,g}],

            refusr_sq:run( [{output, other}],[],
                "mods[.name all_in |m1,m2|](.funs union .funs.calls).name")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}],[],
                "mods(.funs union .funs)")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}],[],
                "mods(.funs union .funs intersect .funs)")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}],[],
                "mods((.funs intersection .funs) union ((.funs minus .funs) union .funs))")
            ),

        ?assertEqual(
            [{list,[]}],

            refusr_sq:run( [{output, other}],[],
                "mods(.funs minus .funs).vars")
            ),

        ?assertEqual(
            [],

            refusr_sq:run( [{output, other}],[],
                "mods(.funs minus .funs)")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods(.funs union .funs)"),

            refusr_sq:run( [{output, other}],[],
                "mods(.funs intersection .funs)")
            ),

        ?assertEqual(
            element(2,hd(refusr_sq:run( [{output, nodes}],[],
                "mods(.funs.called_by union .funs.calls).vars"))),

            element(2,hd(refusr_sq:run( [{output, nodes}],[],
                "(mods.funs.called_by union mods.funs.calls).vars")))
            ),

        ?assertEqual(
            element(2,hd(refusr_sq:run( [{output, nodes}],[],
                "mods(.funs.called_by union .funs.calls).vars"))),

            element(2,hd(refusr_sq:run( [{output, nodes}],[],
                "mods.funs.called_by.vars union mods.funs.calls.vars")))
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}],[],
                "(mods.funs intersection mods.funs)")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs[.name any_in |g,h|]"),

            refusr_sq:run( [{output, other}],[],
                "mods.funs[name=g or name=h]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs[.name any_in mods.funs.name]"),

            refusr_sq:run( [{output, other}],[],
                "mods.funs")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[.funs.name all_in mods.funs.name]"),

            refusr_sq:run( [{output, other}],[],
                "mods")
            ),

        ?assertEqual(
            ordsets:from_list([{chain,[{nopos,"m2:g/1"},{nopos,"m2:g/1"}],"*\n"},
             {chain,[{nopos,"m1:f/1"},{nopos,"m2:h/0"},{nopos,"m1:f/1"}],"*\n"},
             {chain,[{nopos,"m2:h/0"},{nopos,"m1:f/1"},{nopos,"m2:h/0"}],"*\n"}]),

            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods[.name any_in |m1,m2|].funs.(calls union called_by)+"))
            ),

        ?assertEqual(
            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods.funs.(calls union (.called_by minus .called_by))+")),

            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods.funs.(calls)+"))
            ),

        ?assertEqual(
            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods[name=m2].funs[name=g]")),

            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods[name=m2].funs[[true] in .calls U .called_by]"))
            ),

        ?assertEqual(
            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods[name=m2].funs[name=h]")),

            ordsets:from_list(refusr_sq:run( [{output, other}],[],
                "mods[name=m1 or name=m2].funs[.exprs.(sub)+.val all_in (mods[name in |m1,m2|].funs--[true]).exprs.(sub)+.val]"))
            ),

        ?assertEqual(
            run_nodes("mods.funs(.calls.calls U .called_by)"),
            run_nodes("mods.funs I mods(.funs(.calls U .calls).calls U .funs.called_by I mods.funs)")
            )
        ]
    }.

run_nodes(Query) ->
    ordsets:from_list(
        lists:flatten(
            element(2, 
                hd(refusr_sq:run([{output, nodes}], [], Query))))).
    
conv_testfile1() ->
    {module, ?MISC:canonical_filename("m3.erl"),
    "-module(m3).\n"
    "-export([f/0]).\n"
    "f()-> m4:h(ok).\n"
    "stop()-> m4:h(ok).\n"}.
    
conv_testfile2() ->
    {module, ?MISC:canonical_filename("m4.erl"),
    "-module(m4).\n"
    "-export([h/1]).\n"
    "-define(F, stop).\n"
    "h(X)-> ?F.\n"
    "g(stop)-> h;\n"
    "g(st)-> 1;\n"
    "g(X)-> g(X-1).\n"}.

conv_setup() ->
    [add_file(conv_testfile1()),add_file(conv_testfile2())].
conv_cleanup(Files) ->
    ?debugVal([?FileMan,Files]),[?FileMan:drop_file(File)||File<-Files].

conv_test() ->
    {setup,
        conv_setup(),
        fun conv_cleanup/1,
        [
        ?assertEqual(
            [{group_by,{nopos,"m4:g/1"},eq,name,g}],

            refusr_sq:run( [{output, other}],[],
                "mods[.name all_in |m4|](.funs union .funs.calls)[name in |\"g\"|][.body.val any_in .file.funs.name].name")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name=\"m4\"].funs[name='h']"),

            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[.expr.macro_value any_in .file.funs.args.val]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name='m4'].funs[name=\"g\"]"),

            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[.args.val any_in .file.funs.expr.macro_value]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name=m4 or name=m3].funs"),

            refusr_sq:run( [{output, other}],[],
                "mods[.name any_in |'m3', m3, \"m4\"|].funs[arity in |\"0\", 1|]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[name=g]"),

            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[.body.val any_in .arity]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[name in |g, h|]"),

            refusr_sq:run( [{output, other}],[],
                "mods[name=m4].funs[arity in mods.funs.body.val]")
            ),

        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods[name=m3].funs[name in |'stop'|]"),

            refusr_sq:run( [{output, other}],[],
                "mods[name=m3].funs[.name any_in mods.funs.args.val]")
            ),

        ?assertEqual(
            [{group_by,{nopos,"m3.erl"},eq,value,["m3"]},
                {group_by,{nopos,"m4.erl"},eq,value,["m4","X"]}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m3, m4|](.name U .funs.vars.name)")
            ),

        ?assertEqual(
            run_nodes("mods[name in |m3, m4|](.name U .funs.vars.name)"),
            run_nodes("mods[name in |m3, m4|].name U mods[name in |m3, m4|].funs.vars.name")
            ),

        ?assertEqual(
            run_nodes("mods[name in |m3, m4|](.name U .funs.vars.name)"),
            run_nodes("mods[name in |m3, m4|](.name U mods[name in |m3, m4|].funs.vars.name)")
            )
        ]
    }.

iteration_testfile1() ->
    {module, ?MISC:canonical_filename("it1.erl"),
    "-module(it1).\n"
    "-export([f/0]).\n"
    "f()-> it2:h(3).\n"
    "g()-> it2:f().\n"}.
    
iteration_testfile2() ->
    {module, ?MISC:canonical_filename("it2.erl"),
    "-module(it2).\n"
    "-export([f/0, h/1]).\n"
    "-define(F, g).\n"
    "f()-> ?F().\n"
    "g()-> g().\n"
    "h(0)-> 1;\n"
    "h(X)-> h(X-1).\n"}.

iteration_setup() ->
    [add_file(iteration_testfile1()),add_file(iteration_testfile2())].
iteration_cleanup(Files) ->
    [?FileMan:drop_file(File)||File<-Files].

iteration_tests() ->
[
        ?_assertEqual(
            ordsets:from_list(
                [{chain,
                    [{nopos,"it1:f/0"},{nopos,"it2:h/1"},{nopos,"it2:h/1"}],
                    "*\n"},
                 {chain,
                    [{nopos,"it2:f/0"},{nopos,"it2:g/0"},{nopos,"it2:g/0"}],
                    "*\n"}]),

            ordsets:from_list( refusr_sq:run( [{output, other}],[],
                "mods[name in |it1, it2|].funs[name=f].(calls)+") )
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}1"),

            run_nodes("mods.funs.calls")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}2"),

            run_nodes("mods.funs.calls.calls")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}3"),

            run_nodes("mods.funs.calls.calls.calls")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}4"),

            run_nodes("mods.funs.calls.calls.calls.calls")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls.calls}2"),

            run_nodes("mods.funs.calls.calls.calls.calls")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)1"),

            run_nodes("mods.funs(.exprs U .exprs.sub)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)2"),

            run_nodes("mods.funs(.exprs U .exprs(.sub U .sub.sub))")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)3"),

            run_nodes("mods.funs(.exprs U .exprs(.sub U .sub(.sub U .sub.sub)))")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)4"),

            run_nodes("mods.funs(.exprs U .exprs(.sub U .sub(.sub U .sub(.sub U .sub.sub))))")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub.sub)2"),

            run_nodes("mods.funs(.exprs U .exprs(.sub.sub U .sub.sub.sub.sub))")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)+"),

            run_nodes("mods.funs(.exprs U .exprs.(sub)+)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)+"),

            run_nodes("mods.funs(.exprs U .exprs.sub.(sub)+)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub.sub)+"),

            run_nodes("mods.funs(.exprs U .exprs.sub.sub.(sub.sub)+)")
            ),

        ?_assert(
            ordsets:is_subset(
                run_nodes("mods.funs.(calls)+.calls"),
                run_nodes("mods.funs.(calls)+"))
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)+"),

            run_nodes("mods.funs.(calls)+.(calls)+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)+"),

            run_nodes("mods.funs.(calls)+.(calls)+.(calls)+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)+"),

            run_nodes("mods.funs.(exprs.(sub)+.funs)+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)+.body"),

            run_nodes("mods.funs.(exprs.(sub)+.funs)+.body")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)2"),

            run_nodes("mods.funs.(exprs.(sub)+.funs)2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}2"),

            run_nodes("mods.funs.{exprs.(sub)+.funs}2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls[name=f])+"),

            run_nodes("mods.funs.(exprs.(sub)+.funs[name=f])+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls[name=f])+"),

            run_nodes("mods(.funs U .funs.calls[name=f].(calls[name=f])+)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls[name=f]}2"),

            run_nodes("mods.funs.calls[name=f].calls[name=f]")
            ),

        ?_assertEqual(
            run_nodes("mods.funs"),

            run_nodes("mods.funs.(calls[false])+")
            ),

        ?_assertEqual(
            [],

            run_nodes("mods.funs[false].(calls)+")
            ),

        ?_assertEqual(
            [],

            run_nodes("mods.funs.{calls[false]}1")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.{calls}4"),

            run_nodes("mods.funs.{calls}2.{calls}2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)4"),

            run_nodes("mods.funs.(calls)2.(calls)2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.(calls)+"),

            run_nodes("mods.funs.(calls)+.(calls)2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.vars"),

            run_nodes("mods.funs.exprs.(sub)+.vars")
            ),

        ?_assertEqual(
            run_nodes("mods.funs[name=f]"),

            run_nodes("mods.funs.(calls)+[name=f]")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.calls.calls[name=f]"),

            run_nodes("mods.funs.{calls}2[name=f]")
            ),

        ?_assertEqual(
            run_nodes("mods.funs[name=f].exprs"),

            run_nodes("mods.funs.(calls)+[name=f].exprs")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.calls.calls[name=f].exprs"),

            run_nodes("mods.funs.{calls}2[name=f].exprs")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs"),

            run_nodes("mods.funs.exprs.(sub)+.top")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.name"),

            run_nodes("mods.funs.(calls)+.name")
            ),

        ?_assertEqual(
            [],

            run_nodes("mods.funs.{[false]}1")
            ),

        ?_assertEqual(
            run_nodes("mods.funs"),

            run_nodes("mods.funs.([false])+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.vars"),

            run_nodes("mods.funs.([false])+.vars")
            ),

        ?_assertEqual(
            [],

            run_nodes("mods.funs.{[false]}1.vars")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub.sub)+"),

            run_nodes("mods.funs.exprs.({sub}2)+")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.sub.sub.sub.sub"),

            run_nodes("mods.funs.exprs.{{sub}2}2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub.sub)+.reach"),

            run_nodes("mods.funs.exprs.({sub}2)+.reach")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.sub.sub.sub.sub.reach"),

            run_nodes("mods.funs.exprs.{{sub}2}2.reach")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)+"),

            run_nodes("mods.funs.exprs.{(sub)+}2")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs.(sub)+.reach"),

            run_nodes("mods.funs.exprs.{(sub)+}2.reach")
            ),

        ?_assertEqual(
            run_nodes("mods.funs.exprs(.(sub)+.reach U .(sub)+.origin)"),

            run_nodes("mods.funs.exprs.(sub)+(.reach U .origin)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs(.{calls}2.calls U .{calls}2.called_by)"),

            run_nodes("mods.funs.{calls}2(.calls U .called_by)")
            ),

        ?_assertEqual(
            run_nodes("mods.funs(.{calls}2.calls U .{calls}2.called_by).vars"),

            run_nodes("mods.funs.{calls}2(.calls U .called_by).vars")
            ),

        ?_assertEqual(
            run_nodes("mods.funs[name=f].(exprs.(sub)+.funs)2"),

            run_nodes("mods.funs[name=f].{exprs.(sub)+(.funs U .fundef)}2")
            )
        ].

iteration_test_() ->
    {setup,
        fun iteration_setup/0,
        fun iteration_cleanup/1,
        {timeout, 30, iteration_tests()}
    }.


group_testfile1() ->
    {module, ?MISC:canonical_filename("m5.erl"),
    "-module(m5).\n"
    "-export([f/1]).\n"
    "f(Var)-> [Var, m6:h()].\n"}.
    
group_testfile2() ->
    {module, ?MISC:canonical_filename("m6.erl"),
    "-module(m6).\n"
    "-export([h/0]).\n"
    "h()-> ok.\n"
    "g(stop)-> ok;\n"
    "g(X)-> g(X-1).\n"}.

group_setup() ->
    [add_file(group_testfile1()),add_file(group_testfile2())].
group_cleanup(Files) ->
    ?debugVal([?FileMan,Files]),[?FileMan:drop_file(File)||File<-Files].

group_test() ->
    {setup,
        group_setup(),
        fun group_cleanup/1,
        [
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},list,[{nopos,"m5:f/1"}]},
            {group_by,{nopos,"m6.erl"},list,[{nopos,"m6:h/0"},{nopos,"m6:g/1"}]}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs")
            ),
        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}, {groupby, 1}],[],
                "mods.funs")
            ),
        ?assertEqual(
            refusr_sq:run( [{output, other}, {groupby, 1}],[],
                "mods.funs"),

            refusr_sq:run( [{output, other}, {groupby, 1}],[],
                "mods(.funs U .funs)")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,name,f},
             {group_by,{nopos,"m6:h/0"},eq,name,h},
             {group_by,{nopos,"m6:g/1"},eq,name,g}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs.name")
            ),
        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs.name"),

            refusr_sq:run( [{output, other}, {groupby, 2}],[],
                "mods.funs.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},list,[{nopos,"Var"}]},
            {group_by,{nopos,"m6:g/1"},list,[{nopos,"X"}]}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs(.vars U .vars)")
            ),
        ?assertEqual(
            refusr_sq:run( [{output, other}, {groupby, 2}],[],
                "mods.funs(.vars U .vars)"),

            refusr_sq:run( [{output, other}],[],
                "mods.funs(.vars U .vars)")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},list,[{nopos,"Var"}]},
             {group_by,{nopos,"m6.erl"},list,[{nopos,"X"}]}],

            refusr_sq:run( [{output, other}, {groupby, 1}],[],
                "mods[name in |m5,m6|].funs(.vars U .vars)")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".vars.name",["Var"]},
             {group_by,{nopos,"m6:h/0"},eq,".vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".vars.name",["X"]}],

            refusr_sq:run( [{output, other}, {groupby, 2}],[],
                "mods[name in |m5,m6|].funs.vars(.name U .name)")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},list,[{nopos,"m5:f/1"}]}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs[name=f]")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},list,[{nopos,"Var"}]},
             {group_by,{nopos,"m6:g/1"},list,[{nopos,"X"}]}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs.vars")
            ),
        ?assertEqual(
            refusr_sq:run( [{output, other}],[],
                "mods.funs.vars"),

            refusr_sq:run( [{output, other}, {groupby, 2}],[],
                "mods.funs.vars")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},list,[{nopos,"Var"}]},
             {group_by,{nopos,"m6.erl"},list,[{nopos,"X"}]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs.vars")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,name,m5},
             {group_by,{nopos,"m6.erl"},eq,name,m6}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,".funs.name",[f]},
             {group_by,{nopos,"m6.erl"},eq,".funs.name",[h,g]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs.name")
            ),
        ?assertEqual(
            [{chain,[{nopos,"m6:h/0"}],"\n"},
             {chain,[{nopos,"m5:f/1"},{nopos,"m6:h/0"}],"\n"},
             {chain,[{nopos,"m6:g/1"},{nopos,"m6:g/1"}],"*\n"}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|].funs.(calls)+")
            ),
        ?assertEqual(
            [{list,[]}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name = 0].funs.vars")
            ),
        ?assertEqual(
            [{list,[]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name = 0].funs.vars")
            ),
        ?assertEqual(
            [{chain,[{nopos,"m6:g/1"},{nopos,"m6:g/1"},{nopos,"m6:g/1"}],"\n"}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m,1,m6|].funs.{calls}2[name=g]")
            ),
        ?assertEqual(
            [{chain,[{nopos,"m6:g/1"},{nopos,"m6:g/1"},{nopos,"m6:g/1"}],"\n"}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m,1,m6|].funs.{calls}2[name=g]")
            ),
        ?assertEqual(
            [{chain,[{nopos,"m6:g/1"},{nopos,"m6:g/1"}],"*\n"}],

            refusr_sq:run( [{output, other}],[],
                "mods[name in |m5,m6|].funs.(calls)+[name=g]")
            ),
        ?assertEqual(
            [{chain,[{nopos,"m6:g/1"},{nopos,"m6:g/1"}],"*\n"}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|].funs.(calls)+[name=g]")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,".funs.vars.name",["Var"]},
             {group_by,{nopos,"m6.erl"},eq,".funs.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".vars.name",["Var"]},
             {group_by,{nopos,"m6:h/0"},eq,".vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|].funs.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,".funs.vars.name",["Var"]},
             {group_by,{nopos,"m6.erl"},eq,".funs.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|](.funs U .funs).vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".vars.name",["Var"]},
             {group_by,{nopos,"m6:h/0"},eq,".vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|](.funs U .funs).vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"Var"},eq,name,"Var"},
             {group_by,{nopos,"X"},eq,name,"X"}],

            refusr_sq:run( [{output, other},{groupby, 3}],[],
                "mods[name in |m5,m6|](.funs U .funs).vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,".funs.{calls}2.vars.name",[]},
             {group_by,{nopos,"m6.erl"},eq,".funs.{calls}2.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs.{calls}2.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".{calls}2.vars.name",[]},
             {group_by,{nopos,"m6:h/0"},eq,".{calls}2.vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".{calls}2.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|].funs.{calls}2.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m6:g/1"},eq,".vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 3}],[],
                "mods[name in |m5,m6|].funs.{calls}2.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"X"},eq,name,"X"}],

            refusr_sq:run( [{output, other},{groupby, 4}],[],
                "mods[name in |m5,m6|].funs.{calls}2.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"Var"},eq,name,"Var"},
             {group_by,{nopos,"X"},eq,name,"X"}],

            refusr_sq:run( [{output, other},{groupby, 4}],[],
                "mods[name in |m5,m6|].funs.(calls)+.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".vars.name",["Var"]},
             {group_by,{nopos,"m6:h/0"},eq,".vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 3}],[],
                "mods[name in |m5,m6|].funs.(calls)+.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5:f/1"},eq,".(calls)+.vars.name",["Var"]},
             {group_by,{nopos,"m6:h/0"},eq,".(calls)+.vars.name",[]},
             {group_by,{nopos,"m6:g/1"},eq,".(calls)+.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 2}],[],
                "mods[name in |m5,m6|].funs.(calls)+.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},eq,".funs.(calls)+.vars.name",["Var"]},
             {group_by,{nopos,"m6.erl"},eq,".funs.(calls)+.vars.name",["X"]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs.(calls)+.vars.name")
            ),
        ?assertEqual(
            [{group_by,{nopos,"m5.erl"},list,[{nopos,"Var"}]}],

            refusr_sq:run( [{output, other},{groupby, 1}],[],
                "mods[name in |m5,m6|].funs[name=f].vars")
            )
        ]
    }.

add_file({Type, Name, Text}) ->
    File = ?ESG:create(#file{path=Name, type=Type, eol={lf,eol},lastmod=now()}),
    ?ESG:insert(?ESG:root(), file, File),
    ?Graph:mklink(File, incl, File),
    ?FileMan:add_text(File, last, Text),
    ?ESG:finalize(),
    File.

%% Initial selectors
init_sel_testfile() ->
    {module, ?MISC:canonical_filename("init_sel_test.erl"),
     "-module(init_sel_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,\n f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

init_sel_test() ->
    Filename = ?MISC:canonical_filename("init_sel_test.erl"),
    {setup,
     add_file(init_sel_testfile()),
     fun ?FileMan:drop_file/1 ,
     [?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 34}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 34},
                      {ask_missing, false}], '@rec')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 40}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 40},
                      {ask_missing, false}], '@field')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 59}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 59},
                      {ask_missing, false}], '@macro')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 79}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 79},
                      {ask_missing, false}], '@fun')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 24}], '@def'),
                   {none, []}),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 161}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 161},
                      {ask_missing, false}], '@rec')),
       ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 114}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 114},
                      {ask_missing, false}], '@macro')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 139}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 139},
                      {ask_missing, false}], '@var')),
      ?assertEqual(refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 74}], '@def'),
                   refusr_sq_lib:init_sel(
                     [{file, Filename}, {position, 74},
                      {ask_missing, false}], '@mod'))%,
%      ?assertEqual(refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 116}], '@def'),
%                   refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 116},
%                      {ask_missing, false}], '@fun')),
%      ?assertEqual(refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 144}], '@def'),
%                   refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 144},
%                      {ask_missing, false}], '@field')),
%      ?assertEqual(refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 65}], '@def'),
%                   refusr_sq_lib:init_sel(
%                     [{file, Filename}, {position, 65},
%                      {ask_missing, false}], '@fun'))
    ]}.


file_testfile() ->
    {module, ?MISC:canonical_filename("file_test.erl"),
     "-module(file_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

file_test() ->
    Filename = ?MISC:canonical_filename("file_test.erl"),
    {setup,
     add_file(file_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file"),
                   [{list,
                     [{{Filename,1,1},"file_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs"),
                   [{group_by,{nopos,"file_test.erl"},
                    list,
                     [{{Filename,61,97},"file_test:f/1"},
                      {{Filename,99,119},"file_test:g/1"},
                      {{Filename,121,160},"file_test:h/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs"),
                   [{group_by,{nopos,"file_test.erl"},
                    list,[{{Filename,21,45},"rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros"),
                   [{group_by,{nopos,"file_test.erl"},
                     list,[{{Filename,47,59},"F"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.includes"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     list,
                     [{{Filename,1,1},"file_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.included_by"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     list,
                     [{{Filename,1,1},"file_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.imports"),
                   []),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.exports"),
                   []),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.module"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,module,true}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.header"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                    eq,header,false}]),
%atom vs string???
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.name"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,name,file_test}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.dir"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,dir,filename:dirname(Filename)}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.path"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,path,Filename}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.mod_sum"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,mod_sum,239}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.loc"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,loc,8}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.choc"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,choc,161}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.num_of_fun"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,num_of_fun,3}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.num_of_macros"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,num_of_macros,1}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.num_of_records"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,num_of_records,1}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.included_files"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,included_files,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.imported_modules"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,imported_modules,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.number_of_funpath"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,number_of_funpath,3}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.fun_calls_in"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,fun_calls_in,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.fun_calls_out"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,fun_calls_out,1}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.cohesion"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,cohesion,2}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.otp"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,otp,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.max_app_depth"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,max_app_depth,2}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.max_depth_of_calling"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,max_depth_of_calling,2}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.min_depth_of_calling"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,min_depth_of_calling,1}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.max_depth_of_cases"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,max_depth_of_cases,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.num_of_funclauses"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,num_of_funclauses,3}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.branches_of_recursion"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,branches_of_recursion,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.mccabe"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,mccabe,3}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.num_of_funexpr"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,num_of_funexpr,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.number_of_messpass"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,number_of_messpass,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.fun_return_points"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,fun_return_points,3}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.max_length_of_line"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,max_length_of_line,26}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.avg_length_of_line"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,avg_length_of_line,19}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.no_space_after_comma"),
                   [{group_by,
                     {{Filename,1,1},"file_test.erl"},
                     eq,no_space_after_comma,3}])
     ]}.


fun_testfile() ->
    {module, ?MISC:canonical_filename("fun_test.erl"),
     "-module(fun_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

fun_test() ->
    Filename = ?MISC:canonical_filename("fun_test.erl"),
    {setup,
     add_file(fun_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.refs"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,[{{Filename,108,117},"?F(h(Rec))"}]},
                    {group_by,{nopos,"fun_test:h/1"},
                     list,[{{Filename,111,116},"h(Rec)"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.calls"),
                   [{group_by,{{Filename,60,96},"fun_test:f/1"},
                     list,[{nopos,"io:format/1"}]},
                    {group_by,{nopos,"fun_test:g/1"},
                     list,
                     [{{Filename,60,96},"fun_test:f/1"},
                      {{Filename,120,159},"fun_test:h/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.called_by"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,[{{Filename,98,118},"fun_test:g/1"}]},
                    {group_by,{nopos,"fun_test:h/1"},
                     list,[{{Filename,98,118},"fun_test:g/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.args"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,[{{Filename,62,62},"X"}]},
                    {group_by,{nopos,"fun_test:g/1"},
                     list,[{{Filename,100,102},"Rec"}]},
                    {group_by,{nopos,"fun_test:h/1"},
                     list,[{{Filename,122,124},"Rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==f].body"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,
                     [{{Filename,68,84},"io:format(\"alma\")"},
                      {{Filename,95,95},"X"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==f].exprs"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,
                     [{{Filename,62,62},"X"},
                      {{Filename,68,84},"io:format(\"alma\")"},
                      {{Filename,95,95},"X"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==f].vars"),
                   [{group_by,{nopos,"fun_test:f/1"},
                     list,
                     [{{Filename,62,62},"X"}]}]),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.exported")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,exported,true},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,exported,false},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,exported,false}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.name")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,name,format},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,name,f},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,name,h}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.arity")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,arity,1},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,arity,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,arity,1}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.bif")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,bif,false},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,bif,false},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,bif,false}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.pure")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,pure,false},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,pure,false},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,pure,true}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.defined")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,defined,false},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,defined,true},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,defined,true}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.dirty")),
                   lists:usort(
                     [{group_by,{nopos,"io:format/1"},
                       eq,dirty,true},
                      {group_by,{{Filename,60,96},"fun_test:f/1"},
                       eq,dirty,true},
                      {group_by,{{Filename,120,159},"fun_test:h/1"},
                       eq,dirty,false}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.mod")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,mod,io},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,mod,fun_test},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,mod,fun_test}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.exported")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,exported,true},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,exported,false},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,exported,false}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.loc")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,loc,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,loc,2},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,loc,2}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.choc")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,choc,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,choc,36},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,choc,39}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.max_application_depth")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,max_application_depth,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,max_application_depth,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,max_application_depth,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.max_depth_of_calling")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,max_depth_of_calling,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,max_depth_of_calling,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,max_depth_of_calling,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.max_depth_of_cases")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,max_depth_of_cases,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,max_depth_of_cases,0},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,max_depth_of_cases,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.number_of_funclauses")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,number_of_funclauses,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,number_of_funclauses,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,number_of_funclauses,1}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.branches_of_recursion")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                     eq,branches_of_recursion,0},
                    {group_by,{{Filename,60,96},"fun_test:f/1"},
                     eq,branches_of_recursion,0},
                    {group_by,{{Filename,120,159},"fun_test:h/1"},
                     eq,branches_of_recursion,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.mccabe")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,mccabe,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,mccabe,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,mccabe,1}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.calls_for_function")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,calls_for_function,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,calls_for_function,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,calls_for_function,1}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.calls_from_function")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,calls_from_function,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,calls_from_function,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,calls_from_function,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.number_of_funexpr")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,number_of_funexpr,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,number_of_funexpr,0},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,number_of_funexpr,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.number_of_messpass")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,number_of_messpass,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,number_of_messpass,0},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,number_of_messpass,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.fun_return_points")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,fun_return_points,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,fun_return_points,1},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,fun_return_points,1}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.max_length_of_line")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,max_length_of_line,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,max_length_of_line,26},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,max_length_of_line,21}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.average_length_of_line")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,average_length_of_line,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,average_length_of_line,18},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,average_length_of_line,19}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.no_space_after_comma")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,no_space_after_comma,0},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,no_space_after_comma,0},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,no_space_after_comma,0}])),
      ?assertEqual(lists:usort(
                     refusr_sq:run(
                       [{positions, scalar},{output, other}],
                       [{file, Filename}, {position, 34}],
                       "@file.funs.calls.is_tail_recursive")),
                   lists:usort([{group_by,{nopos,"io:format/1"},
                                 eq,is_tail_recursive,unknown},
                                {group_by,{{Filename,60,96},"fun_test:f/1"},
                                 eq,is_tail_recursive,non_rec},
                                {group_by,{{Filename,120,159},"fun_test:h/1"},
                                 eq,is_tail_recursive,non_rec}]))%,
%      ?assertEqual(lists:usort(
%                     refusr_sq:run(
%                       [{positions, scalar},{output, other}],
%                       [{file, Filename}, {position, 34}],
%                       "@file.funs.calls.fun_sum")),
%                   lists:usort([{group_by,{nopos,"io:format/1"},
%                                 eq,fun_sum,0},
%                                {group_by,{{Filename,60,96},"fun_test:f/1"},
%                                 eq,fun_sum,9},
%                                {group_by,{{Filename,120,159},"fun_test:h/1"},
%                                 eq,fun_sum,9}]))
     ]}.

var_testfile() ->
    {module, ?MISC:canonical_filename("var_test.erl"),
     "-module(var_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

var_test() ->
    Filename = ?MISC:canonical_filename("var_test.erl"),
    {setup,
     add_file(var_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].vars"),
                   [{group_by,{nopos,"var_test:h/1"},
                     list,[{{Filename,122,124},"Rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].vars.refs"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,130,132},"Rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].vars.bindings"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,122,124},"Rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].vars.fundef"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,120,159},"var_test:h/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].vars.name"),
                   [{group_by,{{Filename,122,124},"Rec"},
                     eq,name,"Rec"}])
     ]}.

rec_testfile() ->
    {module, ?MISC:canonical_filename("rec_test.erl"),
     "-module(rec_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

rec_test() ->
    Filename = ?MISC:canonical_filename("rec_test.erl"),
    {setup,
     add_file(rec_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs"),
                   [{group_by,{nopos,"rec_test.erl"},
                     list,[{{Filename,20,44},"rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.refs"),
                   [{group_by,{nopos,"rec"},
                     list,
                     [{{Filename,130,139},"Rec#rec.f2"},
                      {{Filename,152,158},"#rec.f1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.fields"),
                   [{group_by,{nopos,"rec"},
                     list,
                     [{{Filename,20,44},"f1"},
                      {{Filename,20,44},"f2"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.file"),
                   [{group_by,{nopos,"rec"},
                     list,[{{Filename,1,1},"rec_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.name"),
                   [{group_by,{{Filename,20,44},"rec"},
                     eq,name,rec}])
     ]}.

recfield_testfile() ->
    {module, ?MISC:canonical_filename("recfield_test.erl"),
     "-module(recfield_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

recfield_test() ->
    Filename = ?MISC:canonical_filename("recfield_test.erl"),
    {setup,
     add_file(recfield_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.fields.refs"),
                   [{group_by,{nopos,"f1"},
                     list,[{{Filename,162,163},"f1"}]},
                    {group_by,{nopos,"f2"},
                     list,[{{Filename,143,144},"f2"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.fields.rec"),
                   [{group_by,{nopos,"f1"},
                     list,[{{Filename,25,49},"rec"}]},
                    {group_by,{nopos,"f2"},
                     list,[{{Filename,25,49},"rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.fields.file"),
                   [{group_by,{nopos,"f1"},
                     list,[{{Filename,1,1},"recfield_test.erl"}]},
                    {group_by,{nopos,"f2"},
                     list,[{{Filename,1,1},"recfield_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.recs.fields.name"),
                   [{group_by,{{Filename,25,49},"f1"},
                     eq,name,f1},
                    {group_by,{{Filename,25,49},"f2"},
                     eq,name,f2}])
     ]}.

macro_testfile() ->
    {module, ?MISC:canonical_filename("macro_test.erl"),
     "-module(macro_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

macro_test() ->
    Filename = ?MISC:canonical_filename("macro_test.erl"),
    {setup,
     add_file(macro_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros"),
                   [{group_by,{nopos,"macro_test.erl"},
                     list,[{{Filename,48,60},"F"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros.refs"),
                   [{group_by,{nopos,"F"},
                     list,[{{Filename,110,111},"?F"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros.file"),
                   [{group_by,{nopos,"F"},
                     list,[{{Filename,1,1},"macro_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros.name"),
                   [{group_by,{{Filename,48,60},"F"},
                     eq,name,"F"}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros.arity"),
                   [{group_by,{{Filename,48,60},"F"},
                     eq,arity,0}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.macros.const"),
                   [{group_by,{{Filename,48,60},"F"},
                     eq,const,true}])
     ]}.

expr_testfile() ->
    {module, ?MISC:canonical_filename("expr_test.erl"),
     "-module(expr_test).\n"
     "-record(rec, {f1=1, f2}).\n"
     "-define(F,f).\n"
     "f(X) -> io:format(\"alma\"),\n"
     "        X.\n"
     "g(Rec) -> ?F(h(Rec)).\n"
     "h(Rec) -> Rec#rec.f2,\n"
     "          #rec.f1.\n"}.

expr_test() ->
    Filename = ?MISC:canonical_filename("expr_test.erl"),
    {setup,
     add_file(expr_testfile()),
     fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].exprs"),
                   [{group_by,{nopos,"expr_test:h/1"},
                     list,
                     [{{Filename,123,125},"Rec"},
                      {{Filename,131,140},"Rec#rec.f2"},
                      {{Filename,153,159},"#rec.f1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].exprs.fundef"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,121,160},"expr_test:h/1"}]},
                    {group_by,{nopos,"Rec#rec.f2"},
                     list,[{{Filename,121,160},"expr_test:h/1"}]},
                    {group_by,{nopos,"#rec.f1"},
                     list,[{{Filename,121,160},"expr_test:h/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.exprs.funs"),
                   [{group_by,{nopos,"io:format(\"alma\")"},
                     list,[{nopos,"io:format/1"}]},
                    {group_by,{nopos,"?F(h(Rec))"},
                     list,
                     [{{Filename,61,97},"expr_test:f/1"},
                      {{Filename,121,160},"expr_test:h/1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].exprs.vars"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,123,125},"Rec"}]},
                    {group_by,{nopos,"Rec#rec.f2"},
                     list,[{{Filename,123,125},"Rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==h].exprs.records"),
                   [{group_by,{nopos,"Rec#rec.f2"},
                     list,[{{Filename,21,45},"rec"}]},
                    {group_by,{nopos,"#rec.f1"},
                     list,[{{Filename,21,45},"rec"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs.exprs.macros"),
                   [{group_by,{nopos,"?F(h(Rec))"},
                     list,[{{Filename,47,59},"F"}]}]),
       ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==g].exprs.sub"),
                   [{group_by,{{Filename,101,103},"Rec"},
                     list,[{{Filename,101,103},"Rec"}]},
                    {group_by,{{Filename,109,118},"?F(h(Rec))"},
                     list,
                     [{{Filename,109,110},"?F"},
                      {{Filename,112,112},"h"},
                      {{Filename,114,116},"Rec"},
                      {{Filename,113,117},"(Rec)"},
                      {{Filename,112,117},"h(Rec)"},
                      {{Filename,111,118},"(h(Rec))"},
                      {{Filename,109,118},"?F(h(Rec))"}]}]),
       ?assertEqual(refusr_sq:run(
                      [{positions, scalar},{output, other}],
                      [{file, Filename}, {position, 34}],
                      "@file.funs[name==g].exprs.param"),
                    [{group_by,{nopos,"?F(h(Rec))"},
                      list,[{{Filename,112,117},"h(Rec)"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==g].exprs.top"),
                   [{group_by,{{Filename,101,103},"Rec"},
                     list,[{{Filename,101,103},"Rec"}]},
                    {group_by,{{Filename,109,118},"?F(h(Rec))"},
                     list,[{{Filename,109,118},"?F(h(Rec))"}]}]),
%%       ?assertEqual(refusr_sq:run(
%%                   [{positions, scalar},{output, other}],
%%                   [{file, Filename}, {position, 34}],
%%                   "@file.funs[name==g].exprs.reach"),
%%                 []),
%%       ?assertEqual(refusr_sq:run(
%%                   [{positions, scalar},{output, other}],
%%                   [{file, Filename}, {position, 34}],
%%                   "@file.funs[name==g].exprs.origin"),
%%                 []),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 34}],
                     "@file.funs[name==g].exprs.file"),
                   [{group_by,{nopos,"Rec"},
                     list,[{{Filename,1,1},"expr_test.erl"}]},
                    {group_by,{nopos,"?F(h(Rec))"},
                     list,[{{Filename,1,1},"expr_test.erl"}]}])
     ]}.
    
clause_testfile() ->
    {module, ?MISC:canonical_filename("clause_test.erl"),
    "-module(clause_test).\n"
    "fun_with_one_clause() -> 42.\n"
    "fun_with_multi_clause(1) -> 1;\n"
    "fun_with_multi_clause([A]) -> A;\n"
    "fun_with_multi_clause({A,B}) -> A+B;\n"
    "fun_with_multi_clause(_) -> alma,\n"
    "korte,\n"
    "ok.\n"}.

clause_test() ->
    Filename = ?MISC:canonical_filename("clause_test.erl"),
    {setup,
     add_file(clause_testfile()),
    fun(File) -> ?FileMan:drop_file(File) end,
     [?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_one_clause].clause"),
                   [{group_by,{nopos,"clause_test:fun_with_one_clause/0"},
                               list,
                               [{{Filename,23,49},
                                 "clause_test:fun_with_one_clause/0_1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_one_clause].clause[index = 2]"),
                   []),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].clause"),
                   [{group_by,{nopos,"clause_test:fun_with_multi_clause/1"},
                               list,
                               [{{Filename,52,80},
                                 "clause_test:fun_with_multi_clause/1_1"},
                                {{Filename,83,113},
                                 "clause_test:fun_with_multi_clause/1_2"},
                                {{Filename,116,150},
                                 "clause_test:fun_with_multi_clause/1_3"},
                                {{Filename,153,195},
                                 "clause_test:fun_with_multi_clause/1_4"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].clause[index = 1]"),
                   [{group_by,{nopos,"clause_test:fun_with_multi_clause/1"},
                               list,
                               [{{Filename,52,80},
                                 "clause_test:fun_with_multi_clause/1_1"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].clause[index = 3]"),
                   [{group_by,{nopos,"clause_test:fun_with_multi_clause/1"},
                               list,
                               [{{Filename,116,150},
                                 "clause_test:fun_with_multi_clause/1_3"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].clause[index = 4]"),
                   [{group_by,{nopos,"clause_test:fun_with_multi_clause/1"},
                               list,
                               [{{Filename,153,195},
                                 "clause_test:fun_with_multi_clause/1_4"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].clause.module"),
                   [{group_by,{nopos,"clause_test:fun_with_multi_clause/1_1"},
                               list,
                               [{{Filename,1,1},
                                 "clause_test.erl"}]},
                     {group_by,{nopos,"clause_test:fun_with_multi_clause/1_2"},
                               list,
                               [{{Filename,1,1},
                                 "clause_test.erl"}]},
                     {group_by,{nopos,"clause_test:fun_with_multi_clause/1_3"},
                               list,
                               [{{Filename,1,1},
                                 "clause_test.erl"}]},
                     {group_by,{nopos,"clause_test:fun_with_multi_clause/1_4"},
                               list,
                               [{{Filename,1,1},
                                 "clause_test.erl"}]}]),
      ?assertEqual(refusr_sq:run(
                     [{positions, scalar},{output, other}],
                     [{file, Filename}, {position, 1}],
                     "@file.funs[name = fun_with_multi_clause].choc:avg"),
                   [{eq,avg,140.0}])
      ]}.
    
