-module(reftr_rules).

-include_lib("referl_paraphrase/include/paraphrase_global.hrl").
-include_lib("referl_lib/include/lib_export.hrl").

-export([prepare/1]).
-export([test/0]).

prepare(Args) ->
    Name   = proplists:get_value(name,   Args),
    Params = proplists:get_value(params, Args),
    Node   = proplists:get_value(node, Args,
                                 try ?Args:expression(Args)
                                 catch {reflib_error,missing_arg,_} -> undefined
                                 end),
    File = try file(Node) catch _:_ -> undefined end,
    fun() ->
            ?silent(begin
                        rule_interpreter:init(),
                        load_all_rules(),
                        rule_interpreter:execute(Name, Params, Node),
                        rule_interpreter:finalize()
                    end),
            File /= undefined andalso ?Transform:touch(File)
    end.

load_all_rules() ->
    {ok, Filenames} = file:list_dir("lib/rules/src/rules/"),
    [rule_interpreter:load_rules({file, File}) || File <- Filenames, filename:extension(File) == ".refac"].

test() ->
    ri:transform(rules, [{name, introduce_farm}, {params, [{'Nw', 12345}]}]).
    
file(Node)->
    [Form] = case Node of 
		 {'$gn', func, _} -> ?Query:exec(Node, ?Fun:definition());
		 _ -> ?Query:exec(Node, ?Query:seq([?Expr:clause(), ?Clause:form()]))
	     end,
    [File] = ?Query:exec(Form, ?Form:file()),
    File.
