%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

-module(rule_interpreter_lib).
-compile(export_all).

-include_lib("referl_core/src/core.hrl").
-include("rules.hrl").

%% =============================================================================
%% PREDICATES
%%
%% By concept, predicate functions are allowed to be partially defined.
%% Those values on which the function is not defined will cause function_clause
%% error, which should be interpreted as false by the caller.
%% =============================================================================

toplevel(?exprnode) -> ?Query:exec(N, [top]) == [N].

%member(?node, L) -> lists:member(N, L);
member(E, L) -> lists:member(E, L).

pattern({{'$gn',_,_}=N, #expr{role=pattern}}) ->
    ?true.

atom(?exprnode(atom)) ->
    ?true.
var(?exprnode(variable)) ->
    ?true.
list(?exprnode(cons)) ->
    ?true.
app(?exprnode(application)) ->
    ?true.

pure(?exprnode) ->
    [Fun] = ?Query:exec(N, ?Expr:dynfunction()),
    not ?Fun:is_dirty(Fun).

evalsto_list(?node) ->
    N2 = ?Dataflow:reach_1st([N], [back], true),
    lists:all(fun list/1, N2).

evalsto_empty_list(?node) ->
    N2 = ?Dataflow:reach_1st([N], [back], true),
    lists:all(fun nil/1, N2).

evalsto_nonempty_list(?node) ->
    N2 = ?Dataflow:reach_1st([N], [back], true),
    lists:all(fun nonnil/1, N2).

evalsto_oneelement_list(?node) ->
    N2 = ?Dataflow:reach_1st([N], [back], true),
    lists:all(fun oneelement_list/1, N2).

nil(?node) ->
    case ?Graph:data(N) of
        #expr{type=cons} -> ?Graph:path(N, [esub]) == [];
        _ -> false
    end.

oneelement_list(?node) ->
    case ?Graph:data(N) of
        #expr{type=cons} -> length(?Graph:path(N, [esub])) == 1;
        _ -> false
    end.

nonnil(?node) ->
    case ?Graph:data(N) of
        #expr{type=cons} ->
            length(?Graph:path(N, [esub])) == 2;
        #expr{type=infix, value='++'} ->
            [L,_] = ?Graph:path(N, [esub]),
            oneelement_list(L);
        _ ->
            true
    end.

%% =============================================================================
%% CFG-related predicates

%% has_base_case(?exprnode) -> has_base_case(form(N));
%% has_base_case(?formnode) ->
%%     cfg_any(N, fun is_base_case_path/2).

%% recursive(?exprnode) -> recursive(form(N));
%% recursive(?formnode) ->
%%     cfg_any(N, fun is_recursive_path/2).

%% multiply_recursive(?exprnode) -> multiply_recursive(form(N));
%% multiply_recursive(?formnode) ->
%%     cfg_any(N, fun is_divide_and_conquer_style/2).

%% divide_and_conquer_style_deep(?formnode, PrevFunForms) ->
%%     cfg_any(N, PrevFunForms, fun is_divide_and_conquer_style_base/2).

%% %% -----------------------------------------------------------------------------

%% cfg_any(From, Pred) ->
%%     %% Recursive thingies...
%%     cfg_any(From, Pred, _Param=From).

%% cfg_any(From, Pred, Param) ->
%%     Paths = cfg_paths(From),
%%     lists:any(fun(Path) -> Pred(Path, Param) end, Paths).

%% %% -----------------------------------------------------------------------------

%% is_base_case_path(Path, FunForm) ->
%%     not is_recursive_path(Path, FunForm).

%% is_recursive_path(Path, FunForm) ->
%%     [FunNode] = ?Query:exec(FunForm, ?Form:func()),
%%     CalledOnPath = called_functions(Path),
%%     lists:member(FunNode, refpp_benchmark:closure(lists:flatten(CalledOnPath))).

%% is_divide_and_conquer_style(Path, FunForm) ->
%%     is_divide_and_conquer_style_base(Path, [FunForm]).

%% is_divide_and_conquer_style_base(Path, FunForms) ->
%%     CalledOnPath = called_functions(Path),
%%     ClosuresOnPath = [{Group, refpp_benchmark:closure(Group)} || Group <- CalledOnPath],

%%     FunNodes = ?Query:exec(FunForms, ?Form:func()),
%%     Recursive = fun({_Group, Closure}) -> ?MISC:intersect(FunNodes, Closure) /= [] end,

%%     case lists:filter(Recursive, ClosuresOnPath) of
%%         [] -> false;
%%         [{Group, _Closure}] ->
%%             Defs = lists:flatten([?Query:exec(Fun, ?Fun:definition()) || Fun <- Group]),
%%             lists:any(fun(FunDef) ->
%%                               case lists:member(FunDef, FunForms) of
%%                                   true ->
%%                                       false;
%%                                   false ->
%%                                       divide_and_conquer_style_deep(attach_data(FunDef), [FunDef | FunForms])
%%                               end
%%                       end, Defs);
%%         _N ->
%%             true
%%     end.


%% =============================================================================
%% FUNCTIONS
%% =============================================================================

merged_name(OldName) -> list_to_atom(atom_to_list(value(OldName)) ++ "_clauses_merged").

value(?exprnode(atom, V)) -> N, V;
value(V) when is_atom(V) -> V.

func(?exprnode) -> func(form(N));
func(?formnode) -> ?Query:exec(N, [fundef]).
func(?exprnode(_, Nam), A) when is_atom(Nam), is_integer(A) -> N, ?Query:exec(?Query:seq([module], ?Fun:find(Nam,A)));
func(N, A) when is_atom(N), is_integer(A) -> ?Query:exec(?Query:seq([module], ?Fun:find(N,A))).

debug(X) -> ?d(X), true.

clause_type(?clausenode(T)) -> N, T.

parent(?node) ->
    [{_, Parent}|_] = ?Syn:parent(N),
    Parent.

generate_varnames(I) ->
    [{var, 0, list_to_atom("GeneratedVar" ++ integer_to_list(J))} || J <- lists:seq(1,I)].

varnames(Data) ->
    [{var, 0, list_to_atom(Name)} || {{'$gn',variable,_},{variable,Name}} <- Data].

boundVars(Nodes) when is_list(Nodes) ->
    lists:flatmap(fun boundVars/1, Nodes);
boundVars(?exprnode) ->
    ?Query:exec(N, ?Expr:varbinds()).

vars(Nodes) when is_list(Nodes) ->
    lists:flatmap(fun vars/1, Nodes);
vars(?node) ->
    ?Query:exec(N, ?Expr:variables()). %% varrefs?

form(?exprnode) -> form(N);
form(N) ->
    [N2] = ?Syn:get_form(N),
    {N2, ?Graph:data(N2)}.

module(?exprnode) ->
    [M] = ?Query:exec(?Syn:get_file(N), ?File:module()),
    M.

fundef(?formnode) ->
    ?Query:exec1(N, ?Form:func(), oops).

funref(?exprnode) ->
    ?Query:exec1(parent_app(N), ?Expr:dynfunction(), oops). %% TODO: multiple funs referred?

attach_data(Ns) when is_list(Ns) -> [attach_data(N) || N <- Ns];
attach_data({'$gn', _, _} = N)   -> {N, ?Graph:data(N)};
attach_data(X) -> X.

parent_app(Node) ->
    case ?Graph:data(Node) of
        #expr{type=application} -> Node;
        _ ->
            Parents = [P || {_,P} <- ?Syn:parent(Node)],
            hd([parent_app(P) || P <- Parents])
    end.

name(?node) ->
    case ?Graph:data(N) of
        #func{name=Name} -> Name;
        #module{name=Name} -> Name;
        _ -> throw(not_a_fun)
    end.
arity(?node) ->
    case ?Graph:data(N) of
        #func{arity=Arity} -> Arity;
        _ -> throw(not_a_fun)
    end.

intersect(L1, L2) when is_list(L1), is_list(L2) ->
    sets:to_list(sets:intersection(sets:from_list(L1), sets:from_list(L2))).

head_OTHER(?node) ->
    N2 = ?Dataflow:reach_1st([N], []),
    Ns = ?Query:exec(N2, [sel_h]),
    lists:flatmap(fun(N3) -> ?Dataflow:reach_1st([N3], []) end, Ns).

tail_OTHER(?node) ->
    N2 = ?Dataflow:reach_1st([N], []),
    ?Query:exec(N2, [sel_t]).

%% TODO should this be generalized?
protected_head(?node) ->
    P = parent(attach_data(N)),
    ?Expr:role(N) == pattern andalso ?Expr:type(N) == match_expr orelse
        case ?Query:exec(P, [esub]) of
            [_, N] -> ?Expr:role(P) == pattern andalso ?Expr:type(P) == match_expr;
            _ -> false
        end.


-define(TorF(X), try X catch _:_ -> false end).

called_functions(Nodes) when is_list(Nodes) ->
    [?Query:exec(Node, ?Expr:function()) || Node <- Nodes, ?TorF(app(attach_data(Node)))].

%% =============================================================================
%% CFG backend interface
%% =============================================================================

cfg_paths(From) ->
    Tab = ets:new(cfg, [duplicate_bag]),
    {CFGEdges, _Apps} = refsc_cfg_utils:build_cfg(From),
    ets:insert(Tab, CFGEdges),
    Paths = get_paths(Tab, _InitPaths = [[From]], _Ready = []),
    lists:filter(fun valid_path/1, Paths).

get_paths(_Tab, [], Ready) ->
    [lists:reverse(RP) || RP <- Ready];
get_paths(Tab, [ [H | _T] = HPath | TPaths], Ready) ->
    case ets:lookup(Tab, H) of
	[] ->
	    get_paths(Tab, TPaths, [HPath | Ready]);
	List ->
	    Targets = [Target || {_, Target, _Label} <- List],
	    NewPaths =
		[ [Elem | HPath] || Elem <- Targets,
                                    not lists:member(Elem, HPath)],
	    get_paths(Tab, NewPaths ++ TPaths, Ready)
    end.

%% get_paths(_Tab, [], Ready) ->
%%     [lists:reverse(RP) || RP <- Ready];
%% get_paths(Tab, [ [H | _T] = HPath | TPaths], Ready) ->
%%     case ets:lookup(Tab, H) of
%%         [] ->
%%             get_paths(Tab, TPaths, [HPath | Ready]);
%%         List ->
%%             {Returns, Targets} = assort_targets(List),
%%             NewPaths    = [ [Elem | HPath] || Elem <- Targets, not lists:member(Elem, HPath)],
%%             NewRetPaths = [ [Elem | HPath] || Elem <- Returns],
%%             RPaths = ret_paths(Tab, NewRetPaths, []),
%%             get_paths(Tab, NewPaths ++ TPaths, RPaths ++ Ready)
%%     end.

%% ret_paths(_Tab, [], Ready) ->
%%     [lists:reverse(RP) || RP <- Ready];
%% ret_paths(Tab,  [[H | _T] = HPath | TPaths], Ready) ->
%%     case ets:lookup(Tab, H) of
%%         [] ->
%%             ret_paths(Tab, TPaths, [HPath | Ready]);
%%         List ->
%%             case assort_targets(List) of
%%                 {[], _Targets} ->
%%                     RPaths = get_paths(Tab, [HPath], []),
%%                     ret_paths(Tab, TPaths, RPaths ++ Ready);
%%                 {Returns, _} ->
%%                     NewRetPaths = [ [Elem | HPath] || Elem <- Returns],
%%                     ret_paths(Tab, NewRetPaths ++ TPaths, Ready)
%%             end
%%     end.

assort_targets(List) ->
    lists:splitwith(fun({_,_,L}) -> L == ret orelse L == new_val end, List).

valid_path([]) -> true;
valid_path(Path) ->
    case lists:last(Path) of
        {error, _} -> false;
        _          -> true
    end.

return_points(?exprnode) ->
    return_points(form(N));
return_points(?formnode) ->
    lists:map(fun return_point/1, cfg_paths(N)).

return_point(Path) ->
    [Last | _] = lists:dropwhile(fun(Node) -> element(1, Node) == ret end, lists:reverse(Path)),
    Last.
%%    case lists:reverse(Path) of
%%        [{ret, _}, Elem | _] -> Elem;
%%        _ -> no_return_point
%%    end.
