%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.

%%% @author Dniel Horpcsi <daniel-h@elte.hu>

Nonterminals

start
directives
directive
rules
rrule
signature
rewrite
rewrite_seq
wherec
conditions
formula
type
controlpath controlpath2 exclincl
app expr argument arguments token tokens.

Terminals

  atom float integer reserved_symbol reserved_word string char var
  '->' '(' ')' '{' '}' '[' ']' ',' '!' '~' '\\' '=' '|'
  '>' '<' '=<' '>=' '/=' 'not' 'and' 'or' 'implies'
  '+' '-' '*' '/'

  '--' ':'

%% keywords
  'uses' 'rule' 'refactoring' 'apply' 'then' 'also' 'ortry' 'in' 'on' 'where' 'globally' 'onsuccess' 'simultanously'
  'exists' 'forall' 'path' 'including' 'excluding' 'satisfies'.

Rootsymbol start.

Unary 999 'in'.
Unary 999 'on'.

Right 500 'also'.
Right 499 'then'.
Right 499 'ortry'.

Left 400 '*'.
Left 400 '/'.
Left 300 '+'.
Left 300 '-'.
Left 200 'and'.
Left 100 'or'.
Left 50 'implies'.

start -> directives rules : {'$1', '$2'}.

directives -> '$empty' : [].
directives -> directive directives : ['$1'|'$2'].
directive -> 'uses' atom : {uses, '$2'}.

rules -> rrule : ['$1'].
rules -> rrule rules : ['$1'|'$2'].

signature -> atom '(' arguments ')' : {'$1', '$3'}.

rrule -> 'rule'        signature rewrite : #rule{signature='$2', rewrite='$3'}.
rrule -> 'refactoring' signature rewrite : #rule{signature='$2', rewrite='$3'}.

rewrite -> 'apply' app              : #rewrite{apply='$2'}.
rewrite -> tokens            wherec : #rewrite{left ='$1', conditions='$2', type=match_only}.
rewrite -> tokens '~'        wherec : #rewrite{left ='$1', conditions='$3', type=remove}.
rewrite ->        '~' tokens wherec : #rewrite{right='$2', conditions='$3', type=insert}.
rewrite -> tokens '~' tokens wherec : #rewrite{left ='$1', right='$3', conditions='$4'}.

rewrite -> '\\' '{' rewrite_seq '\\' '}' : #rewrite{apply='$3'}.

rewrite -> rewrite 'then'  rewrite : ('$1')#rewrite{next={then, '$3'}}.
rewrite -> rewrite 'also'  rewrite : ('$1')#rewrite{next={also, '$3'}}.

rewrite -> 'globally' rewrite : addmodifier('$2', globally).
rewrite -> 'onsuccess' rewrite : addmodifier('$2', onsuccess).
rewrite -> 'simultanously' rewrite : addmodifier('$2', simultanously).
rewrite -> 'in' var rewrite : ('$3')#rewrite{in = '$2'}.
rewrite -> 'on' var rewrite : ('$3')#rewrite{on = '$2'}.

rewrite_seq -> rewrite : ['$1'].
rewrite_seq -> rewrite 'ortry' rewrite_seq : ['$1'|'$3']. 

wherec -> '$empty' : [].
wherec -> 'where'  conditions : '$2'.

conditions -> formula : ['$1'].
conditions -> formula ',' conditions : ['$1'|'$3'].

formula -> atom '(' arguments ')' : {predicate, '$1', '$3'}.
formula -> expr '=' expr : {'=', '$1', '$3'}.
formula -> expr '/=' expr : {'/=', '$1', '$3'}.
formula -> expr '<' expr : {'<', '$1', '$3'}.
formula -> expr '>' expr : {'>', '$1', '$3'}.
formula -> expr '>=' expr : {'>=', '$1', '$3'}.
formula -> expr '=<' expr : {'=<', '$1', '$3'}.
formula -> 'not' formula : {'not', '$2'}.
formula -> formula 'and' formula : {'and', '$1', '$3'}.
formula -> formula 'or' formula : {'or', '$1', '$3'}.
formula -> formula 'implies' formula : {implies, '$1', '$3'}.
formula -> 'exists' 'path' controlpath exclincl : {exists, path, '$3', '$4', []}.
formula -> 'exists' 'path' controlpath exclincl 'satisfies' formula : {exists, path, '$3', '$4', '$6'}.
formula -> 'forall' 'path' controlpath exclincl 'satisfies' formula : {forall, path, '$3', '$4', '$6'}.
formula -> 'exists' type var ':' formula : {exists, '$2', '$1', '$4'}.
formula -> 'forall' type var ':' formula : {forall, '$2', '$1', '$4'}.
formula -> expr '->' expr : {control, '$1', '$3'}.
formula -> expr '~' '>' expr : {data, '$1', '$4'}.
formula -> expr '~' '~' '>' expr : {data, '$1', '$5'}.
formula -> expr '|' '~' '>' expr : {data, '$1', '$5'}.
formula -> '[' formula ']' : {paren, '$2'}.

type -> atom : '$1'.
type -> var  : '$1'. %% ajajjj

controlpath -> expr '--' '>' expr controlpath2 : {controlflow, ['$1', '$4' | '$5']}.
controlpath2 -> '$empty' : [].
controlpath2 -> '--' '>' expr controlpath2 : ['$3'|'$4'].

exclincl -> '$empty' : [].
exclincl -> 'including' expr exclincl : [{including, '$2'} | '$3'].
exclincl -> 'excluding' expr exclincl : [{excluding, '$2'} | '$3'].

app -> atom '(' arguments ')' : {app, '$1', '$3'}.

expr -> atom : '$1'.
expr -> var : '$1'.
expr -> app : '$1'.
expr -> integer : '$1'.
expr -> string : '$1'.
expr -> expr '+' expr : {'+', '$1', '$3'}.
expr -> expr '-' expr : {'-', '$1', '$3'}.
expr -> expr '*' expr : {'*', '$1', '$3'}.
expr -> expr '/' expr : {'/', '$1', '$3'}.
expr -> '\\' tokens '\\' : {code, '$2'}.
expr -> '(' expr ')' : {paren, '$2'}.

arguments -> '$empty' : [].
arguments -> argument : ['$1'].
arguments -> argument ',' arguments : ['$1' | '$3'].

argument -> var : '$1'.
argument -> atom : '$1'.
argument -> float : '$1'.
argument -> integer : '$1'.
argument -> string : '$1'.
argument -> char : '$1'.
argument -> app : '$1'.

tokens -> token : ['$1'].
tokens -> token tokens : ['$1' | '$2'].

token -> var : '$1'.
token -> atom : '$1'.
token -> float : '$1'.
token -> integer : '$1'.
token -> string : '$1'.
token -> char : '$1'.
token -> reserved_symbol : {value_of('$1'), line_of('$1')}.
token -> reserved_word : {value_of('$1'), line_of('$1')}.

%% THESE ARE MARKED AS SPEC IN THE SCANNER

token -> '->'      : {'->', line_of('$1')}.
token -> '('       : {'(', line_of('$1')}.
token -> ')'       : {')', line_of('$1')}.
token -> '{'       : {'{', line_of('$1')}.
token -> '}'       : {'}', line_of('$1')}.
token -> '['       : {'[', line_of('$1')}.
token -> ']'       : {']', line_of('$1')}.

token -> '-'       : {'-', line_of('$1')}.
token -> '+'       : {'+', line_of('$1')}.
token -> '*'       : {'*', line_of('$1')}.
token -> '/'       : {'/', line_of('$1')}.
token -> '<'       : {'<', line_of('$1')}.
token -> '>'       : {'>', line_of('$1')}.
token -> '=<'      : {'=<', line_of('$1')}.
token -> '>='      : {'>=', line_of('$1')}.
token -> '='       : {'=', line_of('$1')}.
token -> '/='       : {'/=', line_of('$1')}.

token -> '!'       : {'!', line_of('$1')}.
token -> ','       : {',', line_of('$1')}.
token -> '--'      : {'--', line_of('$1')}.
token -> ':'       : {':', line_of('$1')}.
token -> '|'       : {'|', line_of('$1')}.
token -> 'and'     : {'and', line_of('$1')}.
token -> 'or'      : {'or', line_of('$1')}.
token -> 'not'     : {'not', line_of('$1')}.

Erlang code.

-include("rules.hrl").

-record(symbol, {line, name}).

value_of(Token) ->
    element(3, Token).

line_of(Token) ->
    element(2, Token).

addmodifier(#rewrite{modifiers=Ms}=R, X) -> R#rewrite{modifiers=[X|Ms]}.
