-module(rule_tester).
-compile(export_all).

-include_lib("referl_core/src/core.hrl").
-include("rules.hrl").

%-define(ssssh(E), E).
-define(ssssh(E), begin {ok, F} = file:open("./ruletest.log", [write]), G = group_leader(), group_leader(F, self()), E, group_leader(G, self()) end).
-define(tempfile, "TMP.erl").
hr() -> io:format(lists:duplicate(60, $-)), io:nl().

%% =============================================================================

do() ->
    %% file:open("./ruletest.log", [write]),
    {ok, Filenames} = file:list_dir("lib/rules/src/tests/"),
    Tests  = [F || F <-Filenames, filename:extension(F) == ".ruletest" orelse
                                  filename:extension(F) == ".reftest"],
    R = lists:flatten([test(M) || M <- Tests]),
    hr(),
    io:format("~p OK~n", [length(lists:filter(fun(X) -> X==ok end, R))]),
    io:format("~p ERROR~n", [length(lists:filter(fun(X) -> X==error end, R))]).

do(T) ->
    {ok, Filenames} = file:list_dir("lib/rules/src/tests/"),
    Tests  = [F || F <-Filenames,
                   filename:rootname(filename:basename(F)) == atom_to_list(T),
                   filename:extension(F) == ".ruletest" orelse
                       filename:extension(F) == ".reftest"],
    R = lists:flatten([test(M) || M <- Tests]),
    hr(),
    io:format("~p OK~n", [length(lists:filter(fun(X) -> X==ok end, R))]),
    io:format("~p ERROR~n", [length(lists:filter(fun(X) -> X==error end, R))]).

%% =============================================================================

test(M) ->
    io:format("Testing ~p...", [list_to_atom(M)]),
    S = read("lib/rules/src/tests/"++M),
    [R|IOs] = re:split(S, "\\*\\*\\*+", [{return,list}]),
    IO0 = re:split(string:join(IOs, "***"), "\\*\\*\\*+|\\-\\-\\-+", [{return,list}]),
    IO = prepare_io(IO0),
    Res = [test_one_case(I, O, R) || {I,O} <- IO],
    io:nl(),
    Res.

test_one_case(Input, ExpectedOutput, Rule0) ->
    try
        Rule = prepare_rule(Rule0),
        ?ssssh(
           begin
               load_input(Input),
               rule_interpreter:init(),
               load_rule(Rule),
               execute_rule(Rule),
               rule_interpreter:finalize()
           end),
        ActualOutput = ?Syn:flat_text(hd(?Query:exec([file]))),
        T1 = rewrite_tree(parse({forms, ExpectedOutput})),
        T2 = rewrite_tree(parse({forms, ActualOutput})),
        if T1 == T2 ->
                io:format(" OK "),
                ok;
           true ->
                io:format(" DIFF "),
                error
        end
    catch
        unsupported ->
            io:format(" IGN ")
    end.

%% =============================================================================

prepare_io([]) -> [];
prepare_io([_]) -> [];
prepare_io([I,O|Rest]) ->
    I2 = "-module(test)."++I,
    O2 = "-module(test)."++O,
    [{I2,O2}|prepare_io(Rest)].

prepare_rule(R) ->
    try
        case parse({expr, R}) of
            [{call,_,{atom,_,RuleName},Args}] ->
                {EArgs,_Bindings} = erl_eval:expr_list(Args, []),
                {RuleName, EArgs};
            [{tuple, _, [{atom,_,NT}, {integer,_,NI}]}, {call,_,{atom,_,RuleName},Args}] ->
                case refcore_graph:get_dbmod() of
                    refdb_mnesia -> ok;
                    _ -> throw(ezt_kihagyom)
                end,
                {EArgs,_Bindings} = erl_eval:expr_list(Args, []),
                {RuleName, EArgs, {'$gn', NT, NI}};
            _ ->
                R
        end
    catch
        throw:ezt_kihagyom -> throw(unsupported);
        _:_ -> R
    end.

load_input(Input) ->
    save(Input, ?tempfile),
    ri:reset(),
    ri:add(?tempfile).

load_rule(R) when is_tuple(R) ->
    {ok, Filenames} = file:list_dir("lib/rules/src/rules/"),
    [rule_interpreter:load_rules({file, File}) || File <- Filenames, filename:extension(File) == ".refac"];
load_rule(R) ->
    rule_interpreter:load_rules({string, R}).

execute_rule({Name, Args}) ->
    rule_interpreter:execute(Name, Args, undefined);
execute_rule({Name, Args, Node}) ->
    rule_interpreter:execute(Name, Args, Node);
execute_rule(_R) ->
    execute_rule({rule, []}).


rewrite_tree(L) when is_list(L) -> lists:map(fun rewrite_tree/1, L);
rewrite_tree({clauses, Cs}) ->
    {clauses, rewrite_tree(Cs)};
rewrite_tree(T) when is_tuple(T) ->
    [Type,_Line|Rest] = tuple_to_list(T),
    list_to_tuple([Type|[rewrite_tree(R) || R <- Rest]]);
rewrite_tree(X) -> X.

read(FileName) ->
    {ok, Device} = file:open(FileName, [read]),
    try   get_all_lines(Device)
    after file:close(Device)
    end.
get_all_lines(Device) ->
    case io:get_line(Device, "") of
        eof  -> [];
        Line -> Line ++ get_all_lines(Device)
    end.

save(S, F) -> file:write_file(F, S).

parse({forms, S}) ->
    {ok, Tokens, _} = erl_scan:string(S),
    parse_forms(Tokens);
parse({expr, S}) ->
    {ok, Tokens, _} = erl_scan:string(S),
    {ok, Tree} = erl_parse:parse_exprs(Tokens),
    Tree.

parse_forms([]) -> [];
parse_forms(Tokens) ->
    {Ts1,[S|Ts2]} = lists:splitwith(fun({dot,_}) -> false; (_) -> true end, Tokens),
    {ok, F} = erl_parse:parse_form(Ts1++[S]),
    [F | parse_forms(Ts2)].
