-module(matMult).
-export([chunk/3, notChunked/4, decomp/1, recomp/1]).

-define(INT_MAX,2147483647).
-define(RAND_MAX,100).
-define(LCG_A,1664525).
-define(LCG_C,1013904223).

index([H|T], 0) -> H;
index([H|T], N) -> index(T, N-1).

rows(MatrixB) -> MatrixB .

join(Pids) ->
   [ receive {Pid, Result} -> Result end || Pid <- Pids ] .

theSkel(ChunkedRows, Cols) ->
		    skel:run([{farm, [{seq, fun(Y) -> lists:map(fun(X) -> mult_prime2(Y, X) end, Cols) end}], 24}], ChunkedRows),
                    receive
                        {sink_results, Results} -> Results
                    end.

mult(Rows, Cols, Chunk, Nrows) -> Parent = self(),
		    % The_Rows = lists:sublist(Rows, Chunk),

		    ChunkedRows = chunk(Rows, Chunk, Nrows),

                    % {Time, Value} = timer:tc(fun() -> theSkel(ChunkedRows, Cols) end),

		    T1 = erlang:now(),
		    %skel:run([{farm, [{seq, fun(Y) -> lists:map(fun(X) -> mult_prime2(Y, X) end, Cols) end}], 24}], ChunkedRows),

                    skel:run([{map, [{seq, fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end}], fun decomp/1, fun recomp/1}], ChunkedRows),

		    receive
		      {sink_results, Results} -> Results
		    end,

                    % join([ spawn(fun() -> Parent ! {self(), lists:map(fun(X) -> mult_prime2(R , X) end, Cols)} end) || R <- ChunkedRows]),
		    % R1 = receive {Pid, Result} -> Result end,
		    T2 = erlang:now(),
		    Time = timer:now_diff(T2, T1),
  		    io:format("Time ~p~n", [Time]),

		    % T3 = erlang:now(),
                    % R2 = lists:map(fun(X) -> mult_prime2(The_Rows, X) end, Cols),
		    % T4 = erlang:now(),
		    % Time2 = timer:now_diff(T4, T3),
		    % io:format("Time out of a spawn: ~p~n", [Time2]),

		    done.

decomp(Input) -> Input.
  %lists:duplicate(4, Input).

recomp([]) -> [];
recomp([Input|Inputs]) ->
  lists:append(Input,recomp(Inputs)).

chunk([], ChunkSize, Len) -> [];
chunk(List, ChunkSize, Len) ->
 			  case (length(List) < ChunkSize) of
			    true -> [List];
			    false -> Chunk = lists:sublist(List, ChunkSize),
			  	     NewList = lists:sublist(List, ChunkSize+1, Len),
			  	     [ Chunk | chunk(NewList, ChunkSize, Len) ]
		   	   end.


mult_prime2([], C) -> [];
mult_prime2([R|Rows], C) -> [ mult_prime(R, C) | mult_prime2(Rows, C) ].


mult_prime(R, C) -> lists:sum( [ A*B   || {A,B} <- lists:zip(R,C)  ] ) .

cols(MatrixB) -> [ [ index(Row, I) || Row <- MatrixB   ] || I <- lists:seq(0, length(MatrixB)-1) ]  .

randvet(0, _) -> [];
randvet(Ncols, S) ->
  %NewS = (?LCG_A * S + ?LCG_C) rem ?INT_MAX,
  % [NewS rem ?RAND_MAX | randvet(Ncols - 1, NewS)].
  [ 1000 | randvet(Ncols - 1 , S)].

randmat(0, _, _) -> [];
randmat(Nrows, Ncols, S) ->
  [randvet(Ncols, S) | randmat(Nrows - 1, Ncols, S + 1)].

productMat(MatrixA, MatrixB, C, Nrows) ->
  % erlang:system_flag(schedulers_online, X),
  mult(rows(MatrixA), cols(MatrixB), C, Nrows).


randvet2(0) -> [];
randvet2(Ncols) -> [ 1000 | randvet2(Ncols -1) ].

genmat(0, _) -> [];
genmat(Nrows, Ncols) ->
  [ randvet2(Ncols) | genmat(Nrows - 1, Ncols)].


chunked(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).

notChunked(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).

sequential(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).

