-module(matMult2).
-export([chunk/3, notChunked/4, decomp/1, recomp/1]).

-define(INT_MAX,2147483647).
-define(RAND_MAX,100).
-define(LCG_A,1664525).
-define(LCG_C,1013904223).

index([H|T], 0) -> H;
index([H|T], N) -> index(T, N-1).

rows(MatrixB) -> MatrixB .

join(Pids) ->
   [ receive {Pid, Result} -> Result end || Pid <- Pids ] .

theSkel(ChunkedRows, Cols) ->
		    skel:run([{farm, [{seq, fun(Y) -> lists:map(fun(X) -> mult_prime2(Y, X) end, Cols) end}], 24}], ChunkedRows),
                    receive
                        {sink_results, Results} -> Results
                    end.

%mult([],_) -> [];
%mult([R|Rows], Cols) -> [ lists:map(fun(X) -> mult_prime(R, X) end, Cols) | mult(Rows, Cols)  ].

%productMat(MatrixA, MatrixB) ->
%  mult1(rows(MatrixA), cols(MatrixB)).


%productMat(MatrixA, MatrixB) ->
%  skel:run([{farm, [{seq,  fun(R) -> ?MODULE:mult2(R, cols(MatrixB) end}],10}], rows(MatrixA)).


%mult1([],_) -> [];
%mult1([R|Rows], Cols) -> [mult2(R, Cols) | mult1(Rows, Cols)].

%mult1([],_) -> [];
%mult1_par([R|Rows], Cols) -> [skel:run([{farm, [{seq,  fun(C) -> mult_prime(R, C) end}],10}], Cols) | mult1(Rows, Cols)  ].

%mult2(_,[]) -> 1;
%mult2(R, [C|Cols]) -> [ mult_prime(R, C) | mult2(R, Cols)  ].


% [lists:map(fun(X) -> mult_prime(R, X) end, Cols) || R <- Rows ]

% lists:map(fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end, Rows)

% mult(Rows, Cols) -> lists:map(fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end, Rows).

% mult(Rows, Cols) ->
%	Fun2 = fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end,
%	lists:map(Fun2, Rows).

% mult(Rows, Cols, Op) ->
%	Fun2 = fun(R) -> lists:map(fun(X) -> OP(R, X) end, Cols) end,
%	lists:map(Fun2, Rows).



% mult(Rows, Cols) ->
%	skel:run([{map, [{seq, fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end}],
% 			   fun ?MODULE:decomp/1,
%		           fun ?MODULE:recomp/1}], Rows),
%

mult(Rows, Cols, Chunk, Nrows) -> Parent = self(),
		   ChunkedRows = chunk(Rows, Chunk, Nrows),
                   skel:run([{map, [{seq, fun(R) -> lists:map(fun(X) -> mult_prime(R,X) end, Cols) end}],
				   fun decomp/1,
			           fun recomp/1}], ChunkedRows),

		    receive
		      {sink_results, Results} -> Results
		    end.

decomp(Input) -> Input.
  %lists:duplicate(4, Input).

recomp([]) -> [];
recomp([Input|Inputs]) ->
  lists:append(Input,recomp(Inputs)).

chunk([], ChunkSize, Len) -> [];
chunk(List, ChunkSize, Len) ->
 			  case (length(List) < ChunkSize) of
			    true -> [List];
			    false -> Chunk = lists:sublist(List, ChunkSize),
			  	     NewList = lists:sublist(List, ChunkSize+1, Len),
			  	     [ Chunk | chunk(NewList, ChunkSize, Len) ]
		   	   end.


mult_prime2([], C) -> [];
mult_prime2([R|Rows], C) -> [ mult_prime(R, C) | mult_prime2(Rows, C) ].


mult_prime(R, C) -> lists:sum( [ A*B   || {A,B} <- lists:zip(R,C)  ] ) .

cols(MatrixB) -> [ [ index(Row, I) || Row <- MatrixB   ] || I <- lists:seq(0, length(MatrixB)-1) ]  .

randvet(0, _) -> [];
randvet(Ncols, S) ->
  %NewS = (?LCG_A * S + ?LCG_C) rem ?INT_MAX,
  % [NewS rem ?RAND_MAX | randvet(Ncols - 1, NewS)].
  [ 1000 | randvet(Ncols - 1 , S)].

randmat(0, _, _) -> [];
randmat(Nrows, Ncols, S) ->
  [randvet(Ncols, S) | randmat(Nrows - 1, Ncols, S + 1)].

productMat(MatrixA, MatrixB, C, Nrows) ->
  % erlang:system_flag(schedulers_online, X),
  mult(rows(MatrixA), cols(MatrixB), C, Nrows).


randvet2(0) -> [];
randvet2(Ncols) -> [ 1000 | randvet2(Ncols -1) ].

genmat(0, _) -> [];
genmat(Nrows, Ncols) ->
  [ randvet2(Ncols) | genmat(Nrows - 1, Ncols)].


chunked(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).

notChunked(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).

sequential(MatrixA, MatrixB, Nrows, Chunk) ->
  productMat(MatrixA, MatrixB, Chunk, Nrows).


run_all_examples(C) ->
  [run_examples(X, C) || X <- [1,2,4,6,8,10,12,14,16,18,20,22,24]].

run_examples(X, C) ->
  erlang:system_flag(schedulers_online, X),
  MatrixA = genmat(1000,1000),
  MatrixB = genmat(1000, 1000),
  Ntimes = 1,
  % Input = lists:duplicate(1024, xx),
  io:format("------~nRunning Examples on ~p cores. Please be Patient. ~n", [X]),
  % io:format("Sequential: ~p~n", [sk_profile:benchmark(fun ?MODULE:example4/1, [MatrixA, MatrixB, 1000, 1], Ntimes)]),
  io:format("No Chunking: ~p~n", [sk_profile:benchmark(fun notChunked/4, [MatrixA, MatrixB, 1000, C], Ntimes)]),
  % io:format("Example 2: ~p~n", [sk_profile:benchmark(fun ?MODULE:example2/1, [Input], Ntimes)]),
  % io:format("Example 1: ~p~n", [sk_profile:benchmark(fun ?MODULE:example1/1, [Input], Ntimes)]),
  io:format("Done with examples on ~p cores.~n------~n", [X]).

