-module(matrix).
-export([main/2]).

-compile(export_all).

-define(INT_MAX,2147483647).
-define(RAND_MAX,100).
-define(LCG_A,1664525).
-define(LCG_C,1013904223).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

main(Nrows, S) ->
  % [Head | _] = Args,
  % IsBench = string:equal(Head, 'is_bench'),
  % {ok, [Nrows, S]} = io:fread("", "~d~d"),
  MatrixA = randmat(Nrows, Nrows, S),
  MatrixB = randmat(Nrows, Nrows, S),
  T1 = erlang:now(),
  productMat(MatrixA, MatrixB),
  T2 = erlang:now(),
  Time = timer:now_diff(T2, T1),
  io:format("Time: ~p~n", [Time]).

productMat(MatrixA, MatrixB) ->
  mult(rows(MatrixA), cols(MatrixB)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mult([],_) -> [];
mult([R|Rows], Cols) -> [lists:map(fun(X) -> multSum(R, X) end, Cols) | mult(Rows, Cols)].   

%% mult_seq_1 --> map --> map --> mult_prime --> local_A*B
%% 		 farm     farm     reduce        farm

mult_seq_1(Rows, Cols) -> lists:map(fun(R) -> 
					lists:map(fun(C) -> multSum(R, C) end, Cols) 
				    end, Rows).

y(R, Cols) -> lists:map(fun(C) -> multSum(R, C) end, Cols) .

mult_par_1(Rows, Cols) -> 
	skel:do([{farm, [{seq, fun(R) -> 
					lists:map(fun(C) -> multSum(R,C) end, Cols) 
			       end}], 10}], Rows).

mult_par_2([],_) -> [];
mult_par_2([R|Rows], Cols) -> 
	[skel:do([{farm, [{seq, fun(C) -> multSum(R,C) end}],10}], Cols) | mult(Rows, Cols)].   

multSum(R, C) -> lists:sum([A*B   || {A,B} <- lists:zip(R,C)  ] ) . 

x(R,C) -> [A*B   || {A,B} <- lists:zip(R,C)  ].

z(A,B) -> A*B. 

multSum_par_1(R,C)  -> 
	lists:sum(skel:do([{farm, [{seq, fun({A,B}) -> A*B end}],10}], lists:zip(R,C))).

multSum_par_2(R,C)  -> 
	skel:do([{reduce, fun(A, B) -> A*B end, fun id/1}], lists:zip(R,C)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
id(X) -> X.

index([H|T], 0) -> H;
index([H|T], N) -> index(T, N-1).

rows(MatrixB) -> MatrixB .

cols(MatrixB) -> [ [ index(Row, I) || Row <- MatrixB   ] || I <- lists:seq(0, length(MatrixB)-1) ]  .

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

randvet(0, _) -> [];
randvet(Ncols, S) ->
  % NewS = (?LCG_A * S + ?LCG_C) rem ?INT_MAX,
  % [NewS rem ?RAND_MAX | randvet(Ncols - 1, NewS)].
  [ 1000 | randvet(Ncols -1 , S) ].

randmat(0, _, _) -> [];
randmat(Nrows, Ncols, S) ->
  [randvet(Ncols, S) | randmat(Nrows - 1, Ncols, S + 1)].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


