-module(matrix_ex_paper_par).

-compile(export_all).

-spec mult_matrix([[integer()]], [[integer()]]) -> [[integer()]].
mult_matrix([], _) ->
  [];
mult_matrix([R | Rows], Cols) ->
  [ lists:map (
          fun(C) -> scalar_product(R,C) end,
          Cols )
  | mult_matrix(Rows,Cols) ].

-spec scalar_product([integer()], [integer()]) -> integer().
scalar_product(R, C) when is_list(R),is_list(C)-> 
  lists:sum([ mult_scalar(A,B) || 
              {A,B} <- zip(R,C) ]);

scalar_product(_,_) ->
    scalar_product(lists:seq(1,500),lists:seq(1,500)).
    
-spec mult_scalar(integer(), integer()) -> integer().
mult_scalar(A, B) ->
  A*B.

zip([X|R],[Y|C])->
    [{X,Y} | zip(R,C)];
zip([], _) ->
    [];
zip(_, []) ->
    [].

%% ----------------------------------------------------------------------
%% for benchmarking

-define(INT_MAX,2147483647).
-define(RAND_MAX,10).
-define(LCG_A,1664525).
-define(LCG_C,1013904223).

randvet(0, _) -> [];
randvet(Ncols, S) ->
    NewS = (?LCG_A * S + ?LCG_C) rem ?INT_MAX,
    [NewS rem ?RAND_MAX | randvet(Ncols - 1, NewS)].

randmat(0, _, _) -> [];
randmat(Nrows, Ncols, S) ->
  [randvet(Ncols, S) | randmat(Nrows - 1, Ncols, S + 1)].

index([H|T], 0) -> H;
index([H|T], N) -> index(T, N-1).

rows(MatrixB) -> MatrixB.
cols(MatrixB) -> [[index(Row, I) || Row <- MatrixB   ] || I <- lists:seq(0, length(MatrixB)-1) ].

productMat(MatrixA, MatrixB) ->
  mult_matrix(rows(MatrixA), cols(MatrixB)).

productMat(MatrixA, MatrixB, V, Nw1, Nw2, Nw3) ->
  FunName = list_to_atom("mult_matrix_par" ++ V),
  matrix_ex_paper_par:FunName(rows(MatrixA), cols(MatrixB), Nw1, Nw2, Nw3).

measure_and_compare(M1, M2, V, Tseq, Nw1, Nw2, Nw3) ->
    T1 = erlang:now(),
    productMat(M1, M2, V, Nw1, Nw2, Nw3),
    T2 = erlang:now(),
    Time = timer:now_diff(T2, T1),
    io:format("ver: ~p, time: ~p, speedup: ~p\n",[V, Time, Tseq/Time]).

main(Nrows, S) ->
    M1 = randmat(Nrows, Nrows, S),
    M2 = randmat(Nrows, Nrows, S),
    T1 = erlang:now(),
    productMat(M1, M2),
    T2 = erlang:now(),
    Tseq = timer:now_diff(T2, T1),
    io:format("Tseq: ~p~n", [Tseq]),
 measure_and_compare(M1, M2, "4", Tseq, 1, 1, 24),
    measure_and_compare(M1, M2, "1", Tseq, 22, 1, 1),
    measure_and_compare(M1, M2, "2", Tseq, 6, 3, 1),
    measure_and_compare(M1, M2, "2", Tseq, 4, 5, 1),
    measure_and_compare(M1, M2, "2", Tseq, 3, 6, 1),
    measure_and_compare(M1, M2, "2", Tseq, 1, 24, 1).

%% ----------------------------------------------------------------------
%% paralell verisons

scalar_product_par(R, C, Nw) when is_list(R),is_list(C)-> 
  lists:sum(begin
               Fun = fun({A,B}) -> mult_scalar(A,B) end,  
               skel:do([{farm, [{seq, Fun}], Nw}], zip(R,C))
             end
            );

scalar_product_par(_,_, Nw) ->
  scalar_product(lists:seq(1,700),lists:seq(1,700)).

% outer
mult_matrix_par1(Rows0, Cols, Nw1, _Nw2, _Nw3) ->
  Fun = fun(H) ->
              case [H | ok] of
                  [] -> hd([]);
                  [H | ok] ->
                      hd([ lists:map (
                          fun(C) -> scalar_product(H,C) end,
                          Cols )
                        | ok ])
              end
      end,
  skel:do([{farm, [{seq, Fun}], Nw1}], Cols).

% outer + middle
mult_matrix_par2(Rows0, Cols, Nw1, Nw2, _Nw3) ->
  Fun = fun(H) ->
              case [H | ok] of
                  [] -> hd([]);
                  [H | ok] ->
                      hd([ skel:do([{farm, [{seq, fun(C) -> 
                                                    scalar_product(H,C)
                                                  end}], Nw2}], Cols)
                        | ok ])
              end
      end,
  skel:do([{farm, [{seq, Fun}], Nw1}], Cols).

% middle
mult_matrix_par3([], _, _, _, _) ->
  [];
mult_matrix_par3([R | Rows], Cols, Nw1, Nw2, Nw3) ->
  [ skel:do([{farm, [{seq, fun(C) -> 
                              scalar_product(R,C)
                            end}], Nw2}], Cols)
  | mult_matrix_par3(Rows,Cols, Nw1, Nw2, Nw3)].

% middle + inner
mult_matrix_par4([], _, _, _, _) ->
  [];
mult_matrix_par4([R | Rows], Cols, Nw1, Nw2, Nw3) ->
  [ skel:do([{farm, [{seq, fun(C) -> 
                              scalar_product_par(R,C,Nw3)
                            end}], Nw2}], Cols)
  | mult_matrix_par4(Rows,Cols, Nw1, Nw2, Nw3)].

