-module(pattern_candidates).

main(BigList) ->
    [some_computation(X) || X <- BigList],
    [some_computation2(some_computation1(X)) || X <- BigList],

    [some_computation(X) || X <- BigList, filter(X)],
    [some_computation2(some_computation1(X)) || X <- BigList, filter(X)],

    lists:map(fun some_computation/1, BigList),
    lists:map(fun(X) -> some_computation2(some_computation1(X)) end, BigList),

    lists:foreach(fun some_computation/1, BigList),
    lists:foreach(fun(X) -> some_computation2(some_computation1(X)) end, 
                  BigList),

    lists:filter(fun filter/1, BigList),
    lists:filter(fun(X) -> filter2(filter1(X)) end, 
                 BigList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
list_comp2(List1, List2)->
    [some_computation(X, Y) || X <- List1, Y <-List2].
    
list_comp2_tr(List1, List2)->
    NewList = [{X, Y} || X <- List1, Y <-List2],
    [some_computation(X, Y) || {X,Y} <- NewList].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
list_comp3(List)->
    [some_computation(X) || X <- BigList, filter(X)].

list_comp3_tr(List)->
    NewList = [X || X<-List, filter(X)],
    [some_computation(X) || X <- NewList].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
applications(List)->
    map1(fun(X)-> X+1 end, List),
    map2(fun(X)-> X+1 end, List),
    map3(fun(X, N)-> X+N+1 end, 0, List),
    map4(List),
    pipeline(fun(X)-> X+1 end, fun(Y) -> Y*2 end, List),
    list_comp(fun(X)-> X+1 end, List),
    [X+1 || X <- List].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
map1(_, []) ->
    [];
map1(F, [H | T]) ->
    [F(H) | map1(F, T)].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
map2(_, []) ->
    [];
map2(F, [H | T]) ->
    Elem = F(H),
    Rem  = map2(F,T),
    [Elem | Rem].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
map3(_, _, []) ->
    [];
map3(F, N, [H | T]) ->
    [F(H, N) | map3(F, N + 1, T)].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
map4([]) ->
    [];
map4([H | T]) ->
    A = some_mod:some_fun1(H),
    B = some_mod:some_fun2(H),
    C = some_mod:some_other_fun(A, B),
    [C | map4(T)].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
map5(List) ->
    case List of
        [] -> [];
        [H|Tail] -> 
            A = some_mod:some_fun1(H),
	    B = some_mod:some_fun2(H),
	    C = some_mod:some_other_fun(A, B),
	    R = map5(T),
	    [C | R]
    end.


%% map4(List) 

%% [fun(X)-> 
%%          A = some_mod:some_fun1(X),
%%          B = some_mod:some_fun2(X),
%%          C = some_mod:some_other_fun(A, B)
%%  end || X <- List]

%% OR

%% [map4_tr(X) || X <- List]
%% map4_tr(X)->
%%     A = some_mod:some_fun1(X),
%%     B = some_mod:some_fun2(X),
%%     C = some_mod:some_other_fun(A, B).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
pipeline(_, _, []) ->
    [];
pipeline(F, G, [H | T]) ->
    [G(F(H)) | pipeline(F, G, T)].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
list_comp(F, List) ->
    [F(P) || P <- List]. 

%% [apply_func(P, P1, P2, P3, ..., PN) || P <- List]
%% foreach


%[do(X) || X <- List]
%skel:pmap(fun do/1, List)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
embedded_pipe(List)->
    [f(X) || X <- List].

f(E)->
    A = do_comp1(E), 
    B = do_comp2(E,A),
    g(A,B).

g(X, Y) ->    
    math:pow(X,Y).

embedded_pipe_tr(List)->
    [g_tr(f_tr(X)) || X <- List].

f_tr(E)->
    A = do_comp1(E), 
    B = do_comp2(E,A),
    {A,B}.

g_tr({X, Y}) ->    
    math:pow(X,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
map_pipe(L)->
    A = [ff(X) || X <- List],
    [gg(X) || X <- A].
%% A is not used

map_pipe_tr(L)->
    [gg(ff(X)) || X <- List].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
f([H | T])->
    A = c1(H),
    B = c2(A),
    [B | f(List)].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LIST. COMP <-> MAP <-> RECURSIVE FUN <-> RECURSIVE FUN WITH ACC

divisible_by_three_lc(List)->
     [ {E, E rem 3 == 0} || E<- List ].

divisible_by_three_map(List)->
     lists:map(fun(E)->
        {E, E rem 3==0}
     end, List).

divisible_by_three_foldl(List)->
    lists:foldl(fun(E, AccIn) ->
        AccIn ++ [{E, E rem 3==0}]
	end, [], List).


divisible_by_three([])->
    [];
divisible_by_three([H| Tail])->
    [ {H, H rem 3 == 0} | divisible_by_three(Tail) ].

divisible_by_three_acc(List)->
    divisible_by_three_acc0(List, []).

divisible_by_three_acc0([], AccIn)->
    AccIn;
divisible_by_three_acc0([H | Tail], AccIn)->
    divisible_by_three_acc0(Tail, AccIn++[{H, H rem 3 == 0}]).


divisible_by_three_acc_(List)->
    divisible_by_three_acc1(List, []).

divisible_by_three_acc1([], AccIn)->
    lists:reverse(AccIn);
divisible_by_three_acc1([H | Tail], AccIn)->
    divisible_by_three_acc1(Tail, [{H, H rem 3 == 0} | AccIn]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LIST. COMP <-> FILTER o MAP <-> RECURSIVE FUN <-> RECURSIVE FUN WITH ACC    

evens_divisible_by_three_lc(List)->
     [ {E, E rem 3 == 0} || E<- List, E rem 2 == 0 ].

evens_divisible_by_three_filter_map(List)->
     lists:map(fun(E)->
        {E, E rem 3==0}
     end, lists:filter(fun(E)-> E rem 2 == 0 end, List)).

evens_divisible_by_three_foldl(List)->
    lists:foldl(fun(E, AccIn) when (E rem 2 == 0) ->
						AccIn ++ [{E, E rem 3==0}];
				   (	_, AccIn) ->
						AccIn
	end, [], List).

evens_divisible_by_three_rec_c([])->
	[];
evens_divisible_by_three_rec_c([H|T])->
	case H rem 2 of
		0 -> [{H, H rem 3 == 0} | evens_divisible_by_three_rec_c(T) ];
		_ -> evens_divisible_by_three_rec_c(T)
	end.

evens_divisible_by_three_rec_fc([])->
	[];
evens_divisible_by_three_rec_fc([H|T]) when H rem 2 == 0 ->
	[{H, H rem 3 == 0} | evens_divisible_by_three_rec_fc(T) ];
evens_divisible_by_three_rec_fc([_|T])->
	evens_divisible_by_three_rec_fc(T).


evens_divisible_by_three_rec_if([])->
	[];
evens_divisible_by_three_rec_if([H|T])->
	if
		H rem 2 == 0 ->
			[{H, H rem 3 == 0} | evens_divisible_by_three_rec_if(T) ];
		true -> evens_divisible_by_three_rec_if(T)
	end.


% anti-candidate of reduce pattern
foldll()->
    lists:foldl(fun(E, Acc0)->
        case length(Acc0) of
            3 -> [];
            _ -> [E|Acc0]
        end
    end, [], lists:seq(-10,2)).
