%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Eötvös Loránd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>PaRTE application</b>
%%%
%%% @todo doc!

-module(paraphrase).
-behaviour(application).

%% Application callbacks
-export([start/2, stop/1]).

start(_Type, _Args) ->
    ReferlNode = generate_referl_node_name(),
    os:cmd(lists:append([
           filename:join([?REFERL_BASE,"bin/referl"]),
           " -sname ",
           ReferlNode,
           " -base ",
           ?REFERL_BASE,
           " -db nif",
           " -dir ",
           filename:join([?REFERL_BASE,"paraphrase_data"]),
           " -server &"])),
    wait_for_referl(list_to_atom(ReferlNode), 10),
    referl_api:start([list_to_atom(ReferlNode)]),
    %error_logger:tty(false),
    wrangler_sup:start().


stop(_State) ->
    referl_api:stop().

generate_referl_node_name()->
    {A1,A2,A3} = now(),
    random:seed(A1, A2, A3),
    ID = random:uniform(1000),
    lists:append(["refactorerl",%integer_to_list(ID), 
		  "@localhost"]).

wait_for_referl(ReferlNode, 0)->
    case net_adm:ping(ReferlNode) of
        pong -> ok;
        _ -> throw("Failed to connect to RefactorErl.")
    end;
wait_for_referl(ReferlNode, NumOfChances)->
    case net_adm:ping(ReferlNode) of
        pong -> ok;
        _ -> timer:sleep(200),
            wait_for_referl(ReferlNode, NumOfChances-1)
    end.
