%%% -*- coding: latin-1 -*-

%%% This file is part of RefactorErl.
%%%
%%% RefactorErl is free software: you can redistribute it and/or modify
%%% it under the terms of the GNU Lesser General Public License as published
%%% by the Free Software Foundation, either version 3 of the License, or
%%% (at your option) any later version.
%%%
%%% RefactorErl is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU Lesser General Public License for more details.
%%%
%%% You should have received a copy of the GNU Lesser General Public License
%%% along with RefactorErl.  If not, see <http://plc.inf.elte.hu/erlang/>.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the Original Code is Etvs Lornd University.
%%% Portions created  by Etvs Lornd University and ELTE-Soft Ltd.
%%% are Copyright 2007-2013 Etvs Lornd University, ELTE-Soft Ltd.
%%% and Ericsson Hungary. All Rights Reserved.
%%%
%%% @author Viktoria Fordos <f-viktoria@elte.hu>
%%% @doc
%%% <b>PaRTE services</b>
%%%
%%% @todo doc!

-module(paraphrase_services).
-export([start_pattern_discovery/3]).
-export([show_prev_results/0]).

-define(WAIT_ONCE, 200).

% @doc Starts Pattern Discovery and displays the result of Pat. Disco. via PC Browser
% Request arrives from Emacs.
% If "Refactorerl Use Basic Configurations" is switched on 
% then the second parameter is the atom 't', otherwise an empty list is given.
% In the former case, RefactorErl is configured at first (refer to configure_referl/3, 
% and only after that Pattern Discovery can continue. 
% In the latter case, the proper configuration of RefactorErl is assumed, 
% so Pattern Discovery immediately starts.
start_pattern_discovery(emacs, t, [Incls, Appbases, Dirs|Tail]=Options) ->
    try
        configure_referl(Incls, Appbases, Dirs),
        {ok, start_pc_browser()}
    catch
        Cl:E ->
            error_logger:error_report([{error_class, Cl},
                                       {error_exp, E},
                                       {stacktrace, erlang:get_stacktrace()}]),
            {error, "Unexpected error occurred."}
    end;

start_pattern_discovery(emacs, _, _) ->
    try
        {ok, start_pc_browser()}
    catch
        Cl:E ->
            error_logger:error_report([{error_class, Cl},
                                       {error_exp, E},
                                       {stacktrace, erlang:get_stacktrace()}]),
            {error, "Unexpected error occurred."}
    end.

% @doc Starts PC Browser to display the previously stored results.
% If no result is present, then a message is returned instead of starting PC Browser.
show_prev_results()->
    case  has_prev_result() of
        true -> show_prev_results0(); 
        _ -> {ok, "No previous result is present!"}
    end.

has_prev_result()->
    try
        referl_api:request({has_prev_result})
    catch
        Cl:E ->
            error_logger:error_report([{error_class, Cl},
                                       {error_exp, E},
                                       {stacktrace, erlang:get_stacktrace()}]),
            {error, "Unexpected error occurred."}
    end. 

show_prev_results0()->
    try
        {ok, start_pc_browser(show)}
    catch
        Cl:E ->
            error_logger:error_report([{error_class, Cl},
                                       {error_exp, E},
                                       {stacktrace, erlang:get_stacktrace()}]),
            {error, "Unexpected error occurred."}
    end.    

start_pc_browser()->
    start_pc_browser(search_and_show).

start_pc_browser(Request)->
    % start_pc_browser
    case Request of
        search_and_show -> referl_api:request({search_and_show_candidates});
        show ->  referl_api:request({show_candidates})
    end,
    WebsrvPath = wait_for_pc_browser(1200),
    "PC Browser should pop up automatically. If no new browser window appears then open a browser and go to " ++ WebsrvPath ++ " to see the result.".

% @doc Configures RefactorErl by using the user defined settings. These settings 
% can be modified in the "Customize RefactorErl" menu item of the "PaRTE" menu in Emacs.
% The configuration process resets the database of RefactorErl, sets environmental values,
% and adds directories (or files) to the database of RefactorErl.
configure_referl(Incls, Appbases, Dirs)->
    referl_api:request({reset}),
    [referl_api:request({add_env, appbase, AppBase}) || AppBase <- Appbases],
    [referl_api:request({add_env, include, Incl}) || Incl <- Incls],
    [referl_api:request({add, Dir}) || Dir <- Dirs].

wait_for_pc_browser(0.0)->
    case referl_api:urequest(refpp_web_srv_controller, server_info,[]) of
        not_started ->
            throw("Timeout exceed");
        _ -> proplists:get_value(url, referl_api:urequest(refpp_web_services,service,[output]))
    end;

wait_for_pc_browser(Sec)->
    case referl_api:urequest(refpp_web_srv_controller, server_info,[]) of
        not_started ->
            timer:sleep(?WAIT_ONCE),
            wait_for_pc_browser(Sec-(1000/?WAIT_ONCE));
        _ -> proplists:get_value(url, referl_api:urequest(refpp_web_services,service,[output]))
    end.
